// -*- C++ -*-
/*!
 * @file  ObjectPoseEstimation.cpp
 * @brief ModuleDescription
 * @date $Date$
 *
 * $Id$
 */

#include "ObjectPoseEstimation.h"
#include <iostream>
#include <stdio.h>

#include "mesaRangeData.h"
#include "pixel.h"
#include "SwissRangerRTCData.h"


#define RECOG_ELEMENT_NUM 20 //ʂo͂z̗vf

//evf͔ẑǂLq邩
#define ELEM_MODELID 1
#define ELEM_ERRORCODE 5
#define ELEM_PARAMETER 8

//eG[R[h
#define ERROR_FILEOPEN_FILED -3
#define ERROR_MODEL_NOTFOUND -105

//#define DEBUG

//wi菜߂̋臒l
float DistanceThresh;
//Mx臒l
float ConfidenceThresh;
//Lu[V֌W
float Chess_size;
int   Pat_row;
int   Pat_col;

//茋ʏo͗p摜
IplImage *srcObjectImg;

int re_estimate = 0;

//肷郂fID
extern int EstimationModelID;
//setModelIDĂ΂ꂽƂtO
extern int executeEstimation;

extern int getObjectPoseExecute(const int id, const CvPoint3D32f PointCloud[25344], std::vector<Model> &ObjectModelList, 
								std::vector<EstimateParameter>& FinalEstimateResultList, std::vector<std::vector<CvPoint3D32f>>& FinalResultVertexList);

// Module specification
// <rtc-template block="module_spec">
static const char* objectposeestimation_spec[] =
  {
    "implementation_id", "ObjectPoseEstimation",
    "type_name",         "ObjectPoseEstimation",
    "description",       "ModuleDescription",
    "version",           "1.0.0",
    "vendor",            "VenderName",
    "category",          "Category",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
	"conf.default.DistanceThresh","0.9",
	"conf.default.Chess_size","23.0",
	"conf.default.Pat_row","7",
	"conf.default.Pat_col","10",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
ObjectPoseEstimation::ObjectPoseEstimation(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_3DPointCloudIn("3DPointCloud", m_3DPointCloud),
    m_SwissRangerDataIn("SwissRangerData", m_SwissRangerData),
    m_RecognitionResultOut("RecognitionResult", m_RecognitionResult),
    m_MDServicePort("MDService"),
    m_OPEServicePort("OPEService"),
    m_ERVServicePort("ERVService"),
    m_RecognitionServicePort("RecognitionService"),
	m_OPEService(&pPoint3DData, &pSwissRangerData, &gCritSec)
    // </rtc-template>
{
	m_OPEService.setMDInterface(&m_MDService);
	m_OPEService.setERVInterface(&m_ERVService);
}

/*!
 * @brief destructor
 */
ObjectPoseEstimation::~ObjectPoseEstimation()
{
}



RTC::ReturnCode_t ObjectPoseEstimation::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("3DPointCloud", m_3DPointCloudIn);
  addInPort("SwissRangerData", m_SwissRangerDataIn);
  
  // Set OutPort buffer
  addOutPort("RecognitionResult", m_RecognitionResultOut);
  
  // Set service provider to Ports
  m_OPEServicePort.registerProvider("OPEService", "ObjectPoseEstimationRTC::idl::OPEService", m_OPEService);
  m_RecognitionServicePort.registerProvider("RecognitionService", "RecognitionService", m_RecognitionService);
  
  // Set service consumers to Ports
  m_MDServicePort.registerConsumer("MDService", "ModelDefinitionRTC::idl::MDService", m_MDService);
  m_ERVServicePort.registerConsumer("ERVService", "ObjectPoseEstimationRTC::idl::ERVService", m_ERVService);
  
  // Set CORBA Service Ports
  addPort(m_MDServicePort);
  addPort(m_OPEServicePort);
  addPort(m_ERVServicePort);
  addPort(m_RecognitionServicePort);

  bindParameter("DistanceThresh", m_DistanceThresh, "0.9");
  bindParameter("Chess_size", m_Chess_size, "23.0");
  bindParameter("Pat_row", m_Pat_row, "7");
  bindParameter("Pat_col", m_Pat_col, "10");
  
  // </rtc-template>

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t ObjectPoseEstimation::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t ObjectPoseEstimation::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t ObjectPoseEstimation::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t ObjectPoseEstimation::onActivated(RTC::UniqueId ec_id)
{
	printf("activate");
	pPoint3DData = new OPEServiceSVC_impl::Point3DList();
	pSwissRangerData = new SRSensor::idl::TimedSwissRanger();

	m_ConfidenceThresh = 0.0;

	CvSize r_size_zoom = cvSize(WIDTH_VALUE*AMP_NUM, HEIGHT_VALUE*AMP_NUM);

	dstObjectImg = cvCreateImage(r_size_zoom, IPL_DEPTH_8U, 1);
	srcObjectImg = cvCreateImage(r_size_zoom, IPL_DEPTH_8U, 1);

	cvNamedWindow("Objects",CV_WINDOW_AUTOSIZE);
	cvMoveWindow("Objects",640+55,340);
	cvZero(dstObjectImg);
	cvZero(srcObjectImg);

  return RTC::RTC_OK;
}


RTC::ReturnCode_t ObjectPoseEstimation::onDeactivated(RTC::UniqueId ec_id)
{
	delete pPoint3DData;
	delete pSwissRangerData;

	cvReleaseImage(&dstObjectImg);
	cvReleaseImage(&srcObjectImg);

	cvDestroyWindow("Objects");

  return RTC::RTC_OK;
}


RTC::ReturnCode_t ObjectPoseEstimation::onExecute(RTC::UniqueId ec_id)
{

	DistanceThresh = m_DistanceThresh;
	Chess_size = m_Chess_size;
	Pat_row = m_Pat_row;
	Pat_col = m_Pat_col;

	ConfidenceThresh = (65535.0 * m_ConfidenceThresh);

	if(m_3DPointCloudIn.isNew() && m_SwissRangerDataIn.isNew())
	{

		m_3DPointCloudIn.read();
		m_SwissRangerDataIn.read();

		for(int i = 0; i < (m_3DPointCloud.data.width * m_3DPointCloud.data.height); i++)
		{
			pPoint3DData->data[i].data.x = m_3DPointCloud.data.range[i].x;
			pPoint3DData->data[i].data.y = m_3DPointCloud.data.range[i].y;
			pPoint3DData->data[i].data.z = m_3DPointCloud.data.range[i].z;
		}

		for(int h = 0; h < m_3DPointCloud.data.height; h++)
		{
			for(int w = 0; w < m_3DPointCloud.data.width; w++)
			{
				pSwissRangerData->data.amplitude[h][w] = m_SwissRangerData.data.amplitude[h][w];
				pSwissRangerData->data.confidence[h][w] = m_SwissRangerData.data.confidence[h][w];
			}
		}
	}
	else
	{
		//f[^|[g̏oĂȂD
		//Ƃ肠l
		for(int i = 0; i < (WIDTH_VALUE * HEIGHT_VALUE); i++)
		{
			
			pPoint3DData->data[i].data.x = 0.0;
			pPoint3DData->data[i].data.y = 0.0;
			pPoint3DData->data[i].data.z = 0.0;
		}
		
		for(int h = 0; h < HEIGHT_VALUE; h++)
		{
			for(int w = 0; w < WIDTH_VALUE; w++)
				pSwissRangerData->data.amplitude[h][w] = 0.0;
		}
	}

	cvCopy(srcObjectImg, dstObjectImg);	

	cvShowImage("Objects",dstObjectImg);
	cvWaitKey(10);

	//setModelIDĂ΂ꂽ琄Jn
	if(executeEstimation == 1)
	{
		int ModelDataNum = 1;

		m_RecognitionResult.data.length(0);
		m_RecognitionResult.data.length(RECOG_ELEMENT_NUM);

		if(CORBA::is_nil(m_MDService.getObject()) || m_MDService.getObject()->_non_existent())
		{
			fprintf(stderr,"T[rX|[gĂ܂\n");
			executeEstimation = 0;
			for(int i = 0; i < RECOG_ELEMENT_NUM; i++)
				m_RecognitionResult.data[i] = 0.0;

			m_RecognitionResult.data[ELEM_ERRORCODE] = (double)ERROR_MODEL_NOTFOUND;

			m_RecognitionResultOut.write();

			return RTC::RTC_OK;
		}
		else
		{	
			int List_num = 0;			
			//idɍv郂ff[^T
			EstimationModel = m_MDService->getObjectModel(EstimationModelID);

			//f̎ނ~̈ȊOꍇw肵ID̃fȂ
			if(EstimationModel.Model_category != 0 && EstimationModel.Model_category != 1)
			{
				fprintf(stderr, "ALL OBJECT SEARCH \n");

     			IDList = m_MDService->getModelIDList();
				List_num = IDList->idlist.length();			
		
				ModelDataNum = IDList->idlist.length();

				EstimationModel_use_List.resize(IDList->idlist.length());

				for(int i = 0; i < IDList->idlist.length(); i++)
				{
					EstimationModel = m_MDService->getObjectModel(IDList->idlist[i]);

					EstimationModel_use_List[i].Model_id = EstimationModel.Model_id;
					EstimationModel_use_List[i].Model_category = EstimationModel.Model_category;

					EstimationModel_use_List[i].wx = EstimationModel.wx;
					EstimationModel_use_List[i].wy = EstimationModel.wy;
					EstimationModel_use_List[i].wz = EstimationModel.wz;

					EstimationModel_use_List[i].height = EstimationModel.height;
					EstimationModel_use_List[i].radius = EstimationModel.radius;
				}

				//͖őS̒T[hƂȂD
/*				
 *				m_RecognitionResult.data.length(RECOG_ELEMENT_NUM);
 *				fprintf(stderr,"f܂ł\n");
 *				executeEstimation = 0;
 *				for(int i = 0; i < RECOG_ELEMENT_NUM; i++)
 *					m_RecognitionResult.data[i] = 0.0;
 *
 *				m_RecognitionResult.data[ELEM_ERRORCODE] = (double)ERROR_MODEL_NOTFOUND;
 *
 *				m_RecognitionResultOut.write();
 *
 *				return RTC::RTC_OK;
 */	

			}
			else
			{
				EstimationModel_use_List.resize(1);

				EstimationModel_use_List[0].Model_id = EstimationModel.Model_id;
				EstimationModel_use_List[0].Model_category = EstimationModel.Model_category;

				EstimationModel_use_List[0].wx = EstimationModel.wx;
				EstimationModel_use_List[0].wy = EstimationModel.wy;
				EstimationModel_use_List[0].wz = EstimationModel.wz;

				EstimationModel_use_List[0].height = EstimationModel.height;
				EstimationModel_use_List[0].radius = EstimationModel.radius;

				printf("id:%d category:%d wx:%.3f wy:%.3f wz:%.3f height:%.3f radius:%.3f\n",EstimationModel_use.Model_id
					,EstimationModel_use_List[0].Model_category, EstimationModel_use_List[0].wx, EstimationModel_use_List[0].wy, EstimationModel_use_List[0].wz
					,EstimationModel_use_List[0].height, EstimationModel_use_List[0].radius);

			}

		}

		//摜`pR|[lg̃CjVCY
		if(CORBA::is_nil(m_ERVService.getObject()) || m_ERVService.getObject()->_non_existent())
		{
		}
		else
		{
			m_ERVService->initializeSuperImpose();
		}

		for(int count = 0; count < TRY_COUNT2; count++)
		{
			printf("TRY_COUNT:%d\n",count);
			//_Qf[^̍XV
			if(m_3DPointCloudIn.isNew() && m_SwissRangerDataIn.isNew())
			{

				m_3DPointCloudIn.read();
				m_SwissRangerDataIn.read();

				for(int i = 0; i < (m_3DPointCloud.data.width * m_3DPointCloud.data.height); i++)
				{
					PointCloud[i].x = m_3DPointCloud.data.range[i].x;
					PointCloud[i].y = m_3DPointCloud.data.range[i].y;
					PointCloud[i].z = m_3DPointCloud.data.range[i].z;
				}

				for(int h = 0; h < m_3DPointCloud.data.height; h++)
				{
					for(int w = 0; w < m_3DPointCloud.data.width; w++)
					{
						pSwissRangerData->data.confidence[h][w] = m_SwissRangerData.data.confidence[h][w];
					}
				}
			}
			else
			{
				//f[^|[g̏oĂȂD
				//Ƃ肠l
				for(int i = 0; i < (WIDTH_VALUE * HEIGHT_VALUE); i++)
				{
					
					PointCloud[i].x = 0.0;
					PointCloud[i].y = 0.0;
					PointCloud[i].z = 0.0;
				}
			}
			
			IplImage *distanceimg;

			CvSize r_size = cvSize(WIDTH_VALUE, HEIGHT_VALUE);
			distanceimg = cvCreateImage(r_size, IPL_DEPTH_8U, 1);
			
			DrawDistanceImg( &distanceimg);
			cvSaveImage("foreground.png", distanceimg);

			cvReleaseImage(&distanceimg);

ReEstimate:

			if(getObjectPoseExecute(EstimationModelID, PointCloud, EstimationModel_use_List, 
				EstimateResultList, ResultVertexList) != 0 && re_estimate == 0)
			{

				fprintf(stderr,"Ɏs܂\n");
				executeEstimation = 0;

				m_RecognitionResult.data.length(RECOG_ELEMENT_NUM);

				for(int i = 0; i < RECOG_ELEMENT_NUM; i++)
					m_RecognitionResult.data[i] = 0.0;

				m_RecognitionResult.data[ELEM_ERRORCODE] = (double)-102;

				m_RecognitionResultOut.write();

				return RTC::RTC_OK;
			}
			else if(re_estimate == 1)
			{
				re_estimate = 0;
				/*f[^B蒼*/
				if(m_3DPointCloudIn.isNew() && m_SwissRangerDataIn.isNew())
				{

					m_3DPointCloudIn.read();
					m_SwissRangerDataIn.read();

					for(int i = 0; i < (m_3DPointCloud.data.width * m_3DPointCloud.data.height); i++)
					{
						PointCloud[i].x = m_3DPointCloud.data.range[i].x;
						PointCloud[i].y = m_3DPointCloud.data.range[i].y;
						PointCloud[i].z = m_3DPointCloud.data.range[i].z;
					}

					for(int h = 0; h < m_3DPointCloud.data.height; h++)
					{
						for(int w = 0; w < m_3DPointCloud.data.width; w++)
						{
							pSwissRangerData->data.confidence[h][w] = m_SwissRangerData.data.confidence[h][w];
						}
					}
				}
				goto ReEstimate;
			}

			cvCopy(srcObjectImg, dstObjectImg);	

			cvShowImage("Objects",dstObjectImg);
			cvWaitKey(10);


			for(int i = 0; i < EstimateResultList.size(); i++)
			{
				printf("--------------EstimateResult---------------\n");
				printf("ID : %d, x:%.3f, y:%.3f, z:%.3f, tx:%.3f, ty:%.3f, tz:%.3f\n"
					,EstimateResultList[i].id
					,EstimateResultList[i].pose.position.x ,EstimateResultList[i].pose.position.y ,EstimateResultList[i].pose.position.z
					,EstimateResultList[i].pose.orientation.r ,EstimateResultList[i].pose.orientation.p ,EstimateResultList[i].pose.orientation.y);
				printf("-------------------------------------------\n");
			}
#ifdef DEBUG			
			printf("ResultVertex\n");
			for(int i = 0; i < ResultVertex.size(); i++)
				printf("[%d]:%.3f %.3f %.3f\n",i,ResultVertex[i].x, ResultVertex[i].y, ResultVertex[i].z);
			
#endif
		}

		//]sɕϊ

		m_RecognitionResult.data.length(RECOG_ELEMENT_NUM * EstimateResultList.size());

		CommonRecogResult.resize(m_RecognitionResult.data.length());

		for(int i = 0; i < EstimateResultList.size(); i++)
		{
			calcMemberofR(EstimateResultList[i].pose, TempCRecogResult);

			for(int j = 0; j < RECOG_ELEMENT_NUM; j++)
				CommonRecogResult[j + (RECOG_ELEMENT_NUM * i)] = TempCRecogResult[j];
		}

#ifdef DEBUG
		printf("RotationMatrix\n");
		printf("%.3f  %.3f  %.3f  %.3f\n",CommonRecogResult[8] ,CommonRecogResult[9] ,CommonRecogResult[10] ,CommonRecogResult[11]);
		printf("%.3f  %.3f  %.3f  %.3f\n",CommonRecogResult[12] ,CommonRecogResult[13] ,CommonRecogResult[14] ,CommonRecogResult[15]);
		printf("%.3f  %.3f  %.3f  %.3f\n",CommonRecogResult[16] ,CommonRecogResult[17] ,CommonRecogResult[18] ,CommonRecogResult[19]);
#endif
		for(int i = 0; i < m_RecognitionResult.data.length(); i++)
			m_RecognitionResult.data[i] = CommonRecogResult[i];
		
		for(int i = 0; i < EstimateResultList.size(); i++)
			m_RecognitionResult.data[ELEM_MODELID + (RECOG_ELEMENT_NUM * i)] = EstimateResultList[i].id;
	
		m_RecognitionResultOut.write();

		//C[t[摜̕`

		for(int i = 0; i < EstimateResultList.size(); i++)
		{
			points.length(ResultVertexList[i].size());

			for(int j = 0; j < points.length(); j++)
			{
				points[j].x = ResultVertexList[i][j].x;
				points[j].y = ResultVertexList[i][j].y;
				points[j].z = ResultVertexList[i][j].z;
#ifdef DEBUG
			printf("%f %f %f\n",points[i].x, points[i].y, points[i].z);
#endif
			}

			if(CORBA::is_nil(m_ERVService.getObject()) || m_ERVService.getObject()->_non_existent())
			{
			}
			else
			{
				m_ERVService->showVertices(points);
			}

		}


		executeEstimation = 0;
	}

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t ObjectPoseEstimation::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t ObjectPoseEstimation::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t ObjectPoseEstimation::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t ObjectPoseEstimation::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t ObjectPoseEstimation::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

void ObjectPoseEstimation::DrawDistanceImg( IplImage **distanceimg){

	int ind, w, h;
	unsigned char uval;
	double maxz = 0.0;

	for(h = 0; h < HEIGHT_VALUE; h++ )
	{
		for(w = 0; w < WIDTH_VALUE; w++)
			maxz = max(maxz, (double)m_3DPointCloud.data.range[(WIDTH_VALUE * h) + w].z);
	}

	for(h = 0; h < HEIGHT_VALUE; h++)
	{
		for(w = 0; w < WIDTH_VALUE; w++)
		{
			uval = (unsigned char)((double)m_3DPointCloud.data.range[(WIDTH_VALUE * h) + w].z * 255.0 / (maxz + 0.1));

			(*distanceimg)->imageData[(*distanceimg)->widthStep * h + w] = uval;
		}
	}
		
}

/*
 * ʒupp[^]švZ
 */
void ObjectPoseEstimation::calcMemberofR(const RTC::Pose3D pose, std::vector<double>& RecogResult)
{

	//Z聨Y聨X̏Ԃɉ]ĂB

	double sx, cx, sy, cy, sz, cz;

	//IFɂ鐄茋ʂ̃tH[}bgɍ킹

 	RecogResult.resize(RECOG_ELEMENT_NUM);
 
	sx = sin(pose.orientation.r);
	cx = cos(pose.orientation.r);

	sy = sin(pose.orientation.p);
	cy = cos(pose.orientation.p);

	sz = sin(pose.orientation.y);
	cz = cos(pose.orientation.y);

	/*
	| R00  R01  R02 |   
	| R10  R11  R12 | = Rot(x, tx)Rot(y, ty)Rot(z, tz)
	| R20  R21  R22 |
	*/

	/*
	IF
	0:JZbgID
	1:fID
	2:Fԍ
	3:Wnԍi0Œj
	4:]l
	5:G[R[h
	6:\i0Œj
	7:\i0Œj
	8:R00
	9:R01
	10:R02
	11:Tx
	12:R10
	13:R11
	14:R12
	15:Ty
	16:R20
	17:R21
	18:R22
	19:Tz
	*/
	for(int i = 0; i < 8; i++)
		RecogResult[i] = 0.0;

	RecogResult[8] = cy * cz;
	RecogResult[9] = -cy * sz;
	RecogResult[10] = sy;
	RecogResult[11] = pose.position.x;
	RecogResult[12] = cx * sz + sx * sy * cz;
	RecogResult[13] = cx * cz - sx * sy * sz;
	RecogResult[14] = - sx * cy;
	RecogResult[15] = pose.position.y;
	RecogResult[16] = sx * sz - cx * sy * cz;
	RecogResult[17] = sx * cz + cx * sy * sz;
	RecogResult[18] = cx * cy;
	RecogResult[19] = pose.position.z;
}

extern "C"
{
 
  void ObjectPoseEstimationInit(RTC::Manager* manager)
  {
    coil::Properties profile(objectposeestimation_spec);
    manager->registerFactory(profile,
                             RTC::Create<ObjectPoseEstimation>,
                             RTC::Delete<ObjectPoseEstimation>);
  }
  
};


