/*************************************************************************/
/*                                                                       */
/*   geo.c: geometrical routines                                         */
/*                                                                       */
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include <iostream>
#include <fstream>

#include <opencv/cv.h>
#include <opencv/highgui.h>

#include "icp3d.h"
#include "myutil.h"
#include "mesaRangeData.h"

//#include "opencv/cvmath/cvmath.h"

#define HEIGHT_VALUE 144
#define WIDTH_VALUE 176

#define PRINT_POINTS_CORRESPONDENCE

extern int disp_errors, disp_errors2, face_num, similar_side, similar_side_cp;

extern int conv_zahyou_p(parameter *object_pose);

extern float DistanceThresh;

double BoxWX;
double BoxWY;
double BoxWZ;

double CylinderHeight;
double CylinderRadius;

extern int correspid;

/*****************************************************************/
/* @     xNgZixNǧvZ, , Oρj@@@   */
/*****************************************************************/
//΃xNgC邢̓xNg̍̂̌vZ
void sub_vector( point p1, point p2, point *ans)
{
  ans->x = p1.x - p2.x;
  ans->y = p1.y - p2.y;
  ans->z = p1.z - p2.z;
}

// xNg̘a
void add_vector( point p1, point p2, point *ans)
{
  ans->x = p1.x + p2.x;
  ans->y = p1.y + p2.y;
  ans->z = p1.z + p2.z;
}

// xNg̃XJ[{ 
void mult_scalar_vector( double a, point p, point *ans)
{
  ans->x = p.x * a;
  ans->y = p.y * a;
  ans->z = p.z * a;
}

//O
void cross_product( point vec1, point vec2, point *ans )
{
    ans->x = vec1.y * vec2.z - vec1.z * vec2.y;
    ans->y = vec1.z * vec2.x - vec1.x * vec2.z;
    ans->z = vec1.x * vec2.y - vec1.y * vec2.x;
}

//
double dot_product( point vec1, point vec2 )
{
    return ( vec1.x * vec2.x + vec1.y * vec2.y + vec1.z * vec2.z );
}

// 2_Ԃ̃[Nbh
double distance(point p1, point p2)
{
  return ( sqrt(sqr(p1.x - p2.x) + sqr(p1.y - p2.y) + sqr(p1.z - p2.z)));
}

/*****************************************************************/
/*                       􉽊wf                            */
/*****************************************************************/
/*  p`f[^̃AP[g  */
polygon *new_polygon(int lnum)
{
  polygon *pp;

  pp = New(polygon);
  pp->line_num = lnum;
  pp->lines = CNew(line, lnum);

  return(pp);
}

/*@ʑ̃f[^̃AP[g@*/
polytope *new_polytope(int pnum, int lnum)
{
  int		i;
  polytope	*pt;

  pt = New(polytope);
  pt->polygon_num = pnum;
  pt->polygons = CNew(polygon, pnum);

  for(i = 0; i < pnum; i++)
  {
	pt->polygons[i].line_num = lnum;
	pt->polygons[i].lines = CNew(line, lnum);
  }

  return(pt);
}
/*@~f̃AP[g@*/
cylinder *new_cylinder()
{
	cylinder *newc;

	newc = New(cylinder);

	return(newc);
}
/*  ʑ̂̊eʂ̖@vZ  */
/*  ̊֐ polygon ̖@tB[hɒlZbgD */
/* 
 * void calc_normal(polygon *polygonp, double *nx, double *ny, double *nz)
 */
void calc_normal(polygon *polygonp)
{
  /*  ʃf[^͍ɓ_ׂĂimake_polytopej,
      Ȃ킿Ciޕɑ΂EOCƂĂD*/
  //p1, p2, p3 -> _W
  double lgth, x, y, z;
  point p0, p1, p2, vec1, vec2, normal;

  /*
    p0, p1, p2 ̏ԂɕłƂC
    N = (p2 - p1) ~ (p0 - p1)
  */

  p0 = polygonp->lines[0].sp;
  p1 = polygonp->lines[0].ep;
  p2 = polygonp->lines[1].ep;

  sub_vector(p2, p1, &vec1);
  sub_vector(p0, p1, &vec2);

  cross_product( vec1, vec2, &normal); /* O */

  /*  K  */
  lgth = sqrt(pow(normal.x, 2.0) + pow(normal.y, 2.0) + pow(normal.z, 2.0));
  x = normal.x / lgth;
  y = normal.y / lgth;
  z = normal.z / lgth;

  if(x == -0.0)	x = 0.0;
  if(y == -0.0)	y = 0.0;
  if(z == -0.0)	z = 0.0;

  polygonp->nx = x;
  polygonp->ny = y;
  polygonp->nz = z;
}

/********************************************************************/
/*   p`f[^̃Rs[iAP[V͂ȂIj  */
void copy_point(point *src, point *dst)
{
  dst->x = src->x;
  dst->y = src->y;
  dst->z = src->z;
}

void copy_line(line *src, line *dst)
{
  copy_point(&(src->sp), &(dst->sp));
  copy_point(&(src->ep), &(dst->ep));
}

void copy_polygon(polygon *src, polygon *dst)
{
  int	i;

  for (i = 0; i < src->line_num; i++)
    copy_line(&(src->lines[i]), &(dst->lines[i]));

  dst->nx = src->nx;
  dst->ny = src->ny;
  dst->nz = src->nz;
}

void copy_polytope(polytope *src, polytope *dst)
{
  int i;

  for (i = 0;i < src->polygon_num; i++)
	copy_polygon(&(src->polygons[i]), &(dst->polygons[i]));
}
/********************************************************************/
/*   ~f[^̃Rs[  */

void copy_disk(disk *src, disk *dst)
{
  int	i;

  copy_point(&(src->center), &(dst->center));  

  dst->nx = src->nx;
  dst->ny = src->ny;
  dst->nz = src->nz;

  dst->radius = src->radius;
}

void copy_cylinder(cylinder *src, cylinder *dst)
{
  int i;

  for (i = 0;i < 2; i++)
	copy_disk(&(src->disks[i]), &(dst->disks[i]));
}
/********************************************************************/

/*  fp`уf[^Cpf[^̐  */
//fʑ̂̐֕ύX

/*  ʑ̃f[^̐  */
/*
  ݂͒̂ɌD3̎厲͍WixCyCzjɉĂC
  wx, wy, wz ͂ꂼ̔̕̕D
*/
void set_polygon_lines4(polygon *pp,
			double x0, double y0, double z0,
			double x1, double y1, double z1,
			double x2, double y2, double z2,
			double x3, double y3, double z3)
{
  pp->lines[0].sp.x = pp->lines[3].ep.x = x0;
  pp->lines[0].sp.y = pp->lines[3].ep.y = y0;
  pp->lines[0].sp.z = pp->lines[3].ep.z = z0;
  
  pp->lines[1].sp.x = pp->lines[0].ep.x = x1;
  pp->lines[1].sp.y = pp->lines[0].ep.y = y1;
  pp->lines[1].sp.z = pp->lines[0].ep.z = z1;
  
  pp->lines[2].sp.x = pp->lines[1].ep.x = x2;
  pp->lines[2].sp.y = pp->lines[1].ep.y = y2;
  pp->lines[2].sp.z = pp->lines[1].ep.z = z2;
  
  pp->lines[3].sp.x = pp->lines[2].ep.x = x3;
  pp->lines[3].sp.y = pp->lines[2].ep.y = y3;
  pp->lines[3].sp.z = pp->lines[2].ep.z = z3;


  calc_normal(pp);
}
			
polytope *make_polytope(double wx, double wy, double wz)
{
  polytope	*pp;

  pp = new_polytope(6, 4);	/* 6ʁCeʂ͎lp` */

  /*  ʂƂɂ̕ӂ̒lZbgD
      @ɌvZD*/

  /*  0: x̐̕  */
  set_polygon_lines4(&(pp->polygons[0]),
		     BoxWX, -BoxWY, -BoxWZ, BoxWX, BoxWY, -BoxWZ,
		     BoxWX, BoxWY, BoxWZ, BoxWX, -BoxWY, BoxWZ);

  /*  1: y̐̕  */
  set_polygon_lines4(&(pp->polygons[1]),
		     BoxWX, BoxWY, -BoxWZ, -BoxWX, BoxWY, -BoxWZ,
		     -BoxWX, BoxWY, BoxWZ, BoxWX, BoxWY, BoxWZ);

  /*  2: z̐̕  */
  set_polygon_lines4(&(pp->polygons[2]),
		     BoxWX, -BoxWY, BoxWZ, BoxWX, BoxWY, BoxWZ,
		     -BoxWX, BoxWY, BoxWZ, -BoxWX, -BoxWY, BoxWZ);

	/*  3: x̕̕  */
  set_polygon_lines4(&(pp->polygons[3]),
		     -BoxWX, BoxWY, -BoxWZ, -BoxWX, -BoxWY, -BoxWZ,
		     -BoxWX, -BoxWY, BoxWZ, -BoxWX, BoxWY, BoxWZ);

  /*  4: y̕̕  */
  set_polygon_lines4(&(pp->polygons[4]),
		     -BoxWX, -BoxWY, -BoxWZ, BoxWX, -BoxWY, -BoxWZ,
		     BoxWX, -BoxWY, BoxWZ, -BoxWX, -BoxWY, BoxWZ);

  /*  5: z̕̕  */
  set_polygon_lines4(&(pp->polygons[5]),
		     BoxWX, BoxWY, -BoxWZ, BoxWX, -BoxWY, -BoxWZ,
		     -BoxWX, -BoxWY, -BoxWZ, -BoxWX, BoxWY, -BoxWZ);

  return pp;
}
cylinder *make_cylinder(double height, double radius)
{
    cylinder *newc;

	newc = new_cylinder();	/* ~f[^̃AP[g */

	newc->disks[0].radius = newc->disks[1].radius = radius;

	newc->disks[0].center.x = newc->disks[0].center.y = 0.0;
	newc->disks[0].center.z = height;

	newc->disks[1].center.x = newc->disks[1].center.y = 0.0;
	newc->disks[1].center.z = -height;

	newc->disks[0].nx = newc->disks[0].ny = 0.0;
	newc->disks[0].nz = 1.0;

	newc->disks[1].nx = newc->disks[1].ny = 0.0;
	newc->disks[1].nz = -1.0;

	return newc;
}
/**************************************************************/
/*                        Wϊ[`                    */
/**************************************************************/
void R2EulerXYZ(CvMat *R, double *tx, double *ty, double *tz)
{
  if ((R->rows != 3) || (R->cols != 3)){
    fprintf(stderr, "R2EulerXYZ: R should be 3x3 matrix (currently %d x %d)\n",
	    R->rows, R->cols);
    exit(-1);
  }
  
  *tx = atan2(- cvmGet(R, 1, 2), cvmGet(R, 2, 2));
  *tz = atan2(- cvmGet(R, 0, 1), cvmGet(R, 0, 0));

  double cy = sign(cvmGet(R, 2, 2)) * sign(cos(*tx))
    * sqrt(sqr(cvmGet(R, 1, 2)) + sqr(cvmGet(R, 2, 2)));
  *ty = atan2(cvmGet(R, 0, 2), cy);
}

void EulerXYZ2R(double *tx, double *ty, double *tz, CvMat *R)
{
  double sx, cx, sy, cy, sz, cz, A, B, C, D, E, F, G, H, I;

  if ((R->rows != 3) || (R->cols != 3)){
    fprintf(stderr, "EulerXYZ2R: R should be 3x3 matrix (currently %d x %d)\n",
	    R->rows, R->cols);
    exit(-1);
  }
  
  sx = sin(*tx);
  cx = cos(*tx);

  sy = sin(*ty);
  cy = cos(*ty);

  sz = sin(*tz);
  cz = cos(*tz);

  /*
    | A  B  C |   
    | D  E  F | = Rot(x, tx)Rot(y, ty)Rot(z, tz)
    | G  H  I |
  */

  A = cy * cz;
  B = -cy * sz;
  C = sy;
  D = cx * sz + sx * sy * cz;
  E = cx * cz - sx * sy * sz;
  F = - sx * cy;
  G = sx * sz - cx * sy * cz;
  H = sx * cz + cx * sy * sz;
  I = cx * cy;

  cvmSet(R, 0, 0, A);
  cvmSet(R, 0, 1, B);
  cvmSet(R, 0, 2, C);
  cvmSet(R, 1, 0, D);
  cvmSet(R, 1, 1, E);
  cvmSet(R, 1, 2, F);
  cvmSet(R, 2, 0, G);
  cvmSet(R, 2, 1, H);
  cvmSet(R, 2, 2, I);
}

/*
  ΍Wn ZCYCX̎ɉ]CɕsړsD
*/
/*  _̍Wϊ */
void transform_point(point *p, double dx, double dy, double dz,
		     double tx, double ty, double tz)
{

  //Z聨Y聨X̏Ԃɉ]ĂB

  double sx, cx, sy, cy, sz, cz, x, y, z, A, B, C, D, E, F, G, H, I;

  sx = sin(tx);
  cx = cos(tx);

  sy = sin(ty);
  cy = cos(ty);

  sz = sin(tz);
  cz = cos(tz);

  /*
    | A  B  C |   
    | D  E  F | = Rot(x, tx)Rot(y, ty)Rot(z, tz)
    | G  H  I |
  */

  A = cy * cz;
  B = -cy * sz;
  C = sy;
  D = cx * sz + sx * sy * cz;
  E = cx * cz - sx * sy * sz;
  F = - sx * cy;
  G = sx * sz - cx * sy * cz;
  H = sx * cz + cx * sy * sz;
  I = cx * cy;

  x = (A * p->x) + (B * p->y) + (C * p->z) + dx;
  y = (D * p->x) + (E * p->y) + (F * p->z) + dy;
  z = (G * p->x) + (H * p->y) + (I * p->z) + dz;
  
  p->x = x;
  p->y = y;
  p->z = z;
}

/*  ̍Wϊ */
void transform_line(line *lp, double dx, double dy, double dz,
		    double tx, double ty, double tz)
{
  transform_point(&(lp->sp), dx, dy, dz, tx, ty, tz);
  transform_point(&(lp->ep), dx, dy, dz, tx, ty, tz);
}

/*  |S̍Wϊ */
void transform_polygon(polygon *pp, double dx, double dy, double dz, double tx, double ty, double tz)
{
  int		i;
  
  /*  e̍Wϊ  */
  for (i = 0; i < pp->line_num; i++){
    transform_line(&(pp->lines[i]), dx, dy, dz, tx, ty, tz);
  }

  /*  @̍ČvZ  */
  calc_normal(pp);
}

/*  ʑ̂̍Wϊ  */
void transform_polytope(polytope *pp, double dx, double dy, double dz, double tx, double ty, double tz)
{
  int	i;

  for (i = 0; i < POLYGON_NUM; i++)
    transform_polygon(&(pp->polygons[i]), dx, dy, dz, tx, ty, tz);
}
/*  ~̏ʁEʂ̍Wϊ */
void transform_disk(disk *dk, double dx, double dy, double dz, double tx, double ty, double tz)
{
  /*  e̍Wϊ  */

	transform_point(&(dk->center), dx, dy, dz, tx, ty, tz);  

}

/*  ~̍Wϊ  */
void transform_cylinder(cylinder *cy, double dx, double dy, double dz, double tx, double ty, double tz)
{
	int	i;

	point vec_axis;

	for (i = 0; i < 2; i++)
		transform_disk(&(cy->disks[i]), dx, dy, dz, tx, ty, tz);


	/*  @̍ČvZ  */
	sub_vector(cy->disks[0].center, cy->disks[1].center, &vec_axis);

	double norm = sqrt(pow(vec_axis.x, 2.0) + pow(vec_axis.y, 2.0) + pow(vec_axis.z, 2.0));

	vec_axis.x /= norm;
	vec_axis.y /= norm;
	vec_axis.z /= norm;

	cy->disks[0].nx = vec_axis.x;
	cy->disks[0].ny = vec_axis.y;
	cy->disks[0].nz = vec_axis.z;

	cy->disks[1].nx = -1.0 * vec_axis.x;
	cy->disks[1].ny = -1.0 * vec_axis.y;
	cy->disks[1].nz = -1.0 * vec_axis.z;
}

//C[t[`p̓_̍Wϊ
void transform_cylinder_image(point *upper, point *lower, double dx, double dy, double dz, double tx, double ty, double tz)
{
	int i;

	for(i = 0; i < 8; i++)
	{
		transform_point(&(upper[i]), dx, dy, dz, tx, ty, tz);
		transform_point(&(lower[i]), dx, dy, dz, tx, ty, tz);
	}

}
/*****************************/
/*
    ʂƓ_̋vZD
 */
/*****************************/
/*  _p`̒ɂ邩ǂ
    ȋʏɂ邩ǂ𒲂ׂ邽߂ɎgjD */
int is_inside_polygon(polygon *polyp, point foot)
{
  point		center, startp, vecA, vecB, vecC, normal; //dSAӂ̊Jn_AexNg
  double	result, nx, ny, nz, length;
  int 		i;

  //dS߂
  center.x = 0.0; center.y = 0.0; center.z = 0.0;
  for(i = 0; i < polyp->line_num; i++)
  {
    center.x += polyp->lines[i].sp.x;
    center.y += polyp->lines[i].sp.y;
    center.z += polyp->lines[i].sp.z;
  }
  center.x /= polyp->line_num;
  center.y /= polyp->line_num;
  center.z /= polyp->line_num;

  nx = polyp->nx;
  ny = polyp->ny;
  nz = polyp->nz;

  for(i = 0; i < polyp->line_num; i++)
  {
	/*

	̖@𒲂ׂ܂BWxNg B WxNg A ăxNgAB߂܂B
    var abx = bx - ax;
    var aby = by - ay;

	ʂ̖@xNg-̏ꍇAepsp
	+̏ꍇAspep

	nx = aby;
	ny = -abx;
@
	@𐳋K܂B
    var length = Math.sqrt((nx * nx) + (ny * ny));
    if(length > 0)	length = 1 / length;
    nx *= length;
    ny *= length;

	*/

	//@߂
	if(nx < 0.0)
	{
	  startp = polyp->lines[i].sp;

	  sub_vector(polyp->lines[i].ep, startp, &vecA);	//ep-sp

	  normal.x = vecA.x;
	  normal.y = -(vecA.z);
	  normal.z = vecA.y;
	}
	else if(nx > 0.0)
	{	  
	  startp = polyp->lines[i].sp;

	  sub_vector(polyp->lines[i].ep, startp, &vecA);	//ep-sp

	  normal.x = vecA.x;
	  normal.y = -(vecA.z);
	  normal.z = vecA.y;
	}
	else if(ny < 0.0)
	{
	  startp = polyp->lines[i].sp;

	  sub_vector(polyp->lines[i].ep, startp, &vecA);	//ep-sp

	  normal.x = -(vecA.z);
	  normal.y = vecA.y;
	  normal.z = vecA.x;
	}
	else if(ny > 0.0)
	{
	  startp = polyp->lines[i].sp;

	  sub_vector(polyp->lines[i].ep, startp, &vecA);	//ep-sp

	  normal.x = -(vecA.z);
	  normal.y = vecA.y;
	  normal.z = vecA.x;
	}
	else if(nz < 0.0)
	{
	  startp = polyp->lines[i].sp;

	  sub_vector(polyp->lines[i].ep, startp, &vecA);	//ep-sp

	  normal.x = -(vecA.y);
	  normal.y = vecA.x;
	  normal.z = vecA.z;
	}
	else
	{
	  startp = polyp->lines[i].sp;

	  sub_vector(polyp->lines[i].ep, startp, &vecA);	//ep-sp

	  normal.x = -(vecA.y);
	  normal.y = vecA.x;
	  normal.z = vecA.z;
	}
	
	length = sqrt(pow(normal.x,2.0) + pow(normal.y,2.0) + pow(normal.z,2.0));
    
	if(length > 0)	length = 1 / length;
   
	normal.x *= length;
	normal.y *= length;
	normal.z *= length;

	sub_vector(center, startp, &vecB);	//dS-sp
	sub_vector(foot, startp, &vecC);  //̑-sp

	result = dot_product(normal, vecB) * dot_product(normal, vecC);

	if(result < 0.0)
	  return 0;
  }
	
  return 1;
}


double distance_to_the_nearest_line(polygon *poly, point *p)
{
  int		i;
  double	dist, mindist, k;
  double	vv, vps;
  point 	v, s, pMINUSs, x, kv;

  mindist = VERY_LARGE;
  for (i = 0; i < poly->line_num; i++){
    s = poly->lines[i].sp;
    sub_vector(poly->lines[i].ep, poly->lines[i].sp, &v);
    sub_vector(*p, s, &pMINUSs);
    vv = dot_product(v, v);
    vps = dot_product(v, pMINUSs);
    k = vps / vv;
    mult_scalar_vector(k, v, &kv);
    add_vector(s, kv, &x);
    dist = distance(*p, x);
    mindist = min(dist, mindist);
  }

  return(mindist);
}


/*
  _p`̒ɂ邩ǂD
  ȋʏɂ邩ǂ𒲂ׂ邽߂Ɏgj

  ʂ̏dSƂ_ꂼɂāCeӂƂłOp`̒_̏ԂD
 */
int is_inside_polygon2(polygon *poly, point *p)
{
  int	i;
  point	cp;
  point	nc, nf, vec1, vec2;

  cp.x = cp.y = cp.z = 0.0;
  for (i = 0; i < poly->line_num; i++){
    cp.x += poly->lines[i].sp.x;
    cp.y += poly->lines[i].sp.y;
    cp.z += poly->lines[i].sp.z;
  }
  cp.x /= (double)poly->line_num;
  cp.y /= (double)poly->line_num;
  cp.z /= (double)poly->line_num;

  /*
    lines[i].sp 𒆐SClines[i].ep 1xNgCړ_2xNgƂD
  */
  for (i = 0; i < poly->line_num; i++){
    sub_vector(poly->lines[i].ep, poly->lines[i].sp, &vec1);
    sub_vector(cp, poly->lines[i].sp, &vec2);
    cross_product(vec1, vec2, &nc);

    sub_vector(*p, poly->lines[i].sp, &vec2);
    cross_product(vec1, vec2, &nf);

    if (dot_product(nc, nf) < 0.0) return(0);
  }
  return(1);
}
 
/*
  _~̒ɂ邩ǂD
  ȋʏɂ邩ǂ𒲂ׂ邽߂Ɏgj
 */
int is_inside_disk(disk *diskp, point *p)
{
  //dS
  point	cp, pp;

  double dist;

  //dSo

  cp.x = diskp->center.x;
  cp.y = diskp->center.y;
  cp.z = diskp->center.z;

  pp.x = p->x;
  pp.y = p->y;
  pp.z = p->z;

  /*
    lines[i].sp 𒆐SClines[i].ep 1xNgCړ_2xNgƂD
  */
  //ʂ̒C1Ԃ
  //dSƐ̑̋𒲂ׁCaȓʂ̒

  dist = sqrt(pow((pp.x - cp.x),2.0) + pow((pp.y - cp.y),2.0) + pow((pp.z - cp.z),2.0));

  //aȉ->ʂ̒
  if(fabs(dist) <= (diskp->radius))
  {
	  return 1;
  }
  else
  {
	  return 0;
  }
}


/*
  ~Ɠ_Ƃ̋vZD
  _~ɉ낵̑ʏɂȂƂɂ́C傫lԂD
*/
double disk_point_distance(disk *diskp, point *pp)
{
  double t, nx, ny, nz, px, py, pz, x1, y1, z1, d;
  double dist;

  point foot;

  nx = diskp->nx;
  ny = diskp->ny;
  nz = diskp->nz;

  px = pp->x;
  py = pp->y;
  pz = pp->z;

  /*
    nx x + ny y + nz z + d = 0 Ƃʂ̕ d ߂D
  */
  d = -(  (nx * diskp->center.x) + (ny * diskp->center.y) + (nz * diskp->center.z));

  /*
    ʂ܂ł̕tDʂ̊Oi@jɂƕɂȂD
  */
  t = ((-d) - (nx * px) - (ny * py) - (nz * pz))/(sqr(nx) + sqr(ny) + sqr(nz));

  x1 = px + (t * nx);
  y1 = py + (t * ny);
  z1 = pz + (t * nz);

  foot.x = x1; foot.y = y1; foot.z = z1;

  if(! is_inside_disk(diskp, &foot)){
/* 
 *   if(! is_inside_polygon(polyp, foot)){
 */
    dist = VERY_LARGE;
/* 
 *     dist = distance_to_the_nearest_line(polyp, &foot);
 */
  }
  else{
    dist = sqrt(pow(px - x1, 2.0) + pow(py - y1, 2.0) + pow(pz - z1, 2.0));

    /*  Ή_̑gDfobOp  */
    if (disp_errors && disp_errors2)
	{
/* 
 *	     fprintf(datafilep, "%f %f %f\n", px, py, pz);
 *	     fprintf(datafilep, "%f %f %f\n", x1, y1, z1);
 */
	  printf("disk_point_distance\n");

	  fprintf(stderr, "data:(%f, %f, %f) - model:(%f, %f, %f) = %f\n",
		  px, py, pz, x1, y1, z1, dist);
	  fprintf(stderr, "[%d] n:(%f, %f, %f)	     n:(%f, %f, %f)\n", face_num,
		  pp->nx, pp->ny, pp->nz, nx, ny, nz);
	  getchar();
	}
  }

  return(dist);
}

//~̑ʂƓ_Ƃ̋vZD
double side_point_distance(cylinder *cyp, point *pp)
{
  double t, rad;
	  //, nx, ny, nz, px, py, pz, x1, y1, z1, d;
  double dist;

  point sp, ep, crossp, foot, vec_axis, datap, vec_data_axis, viewing_direction;

  viewing_direction.x = viewing_direction.y = 0.0;
  viewing_direction.z = 1.0;

  rad = cyp->disks[0].radius;

  sp = cyp->disks[0].center;
  ep = cyp->disks[1].center;

  vec_axis.x = ep.x - sp.x;
  vec_axis.y = ep.y - sp.y;
  vec_axis.z = ep.z - sp.z;

  double axis_norm = sqrt(pow(vec_axis.x, 2.0) + pow(vec_axis.y, 2.0) + pow(vec_axis.z, 2.0));

  vec_axis.x /= axis_norm;
  vec_axis.y /= axis_norm;
  vec_axis.z /= axis_norm;

  datap.x = pp->x;
  datap.y = pp->y;
  datap.z = pp->z;
  datap.nx = pp->nx;
  datap.ny = pp->ny;
  datap.nz = pp->nz;

  //p[^t𓱏o

  t = dot_product(datap, vec_axis) - dot_product(sp, vec_axis);

  //厲Ƃ̌_xoł

  crossp.x = sp.x + (t * vec_axis.x);
  crossp.y = sp.y + (t * vec_axis.y);
  crossp.z = sp.z + (t * vec_axis.z);

  //厲Ƃ̌_f[^_ւ̃xNgƁCf[^_̖@Ƃ̂Ȃp

  vec_data_axis.x = datap.x - crossp.x;
  vec_data_axis.y = datap.y - crossp.y;
  vec_data_axis.z = datap.z - crossp.z; 

  double norm = sqrt(pow(vec_data_axis.x, 2.0) + pow(vec_data_axis.y, 2.0) + pow(vec_data_axis.z, 2.0));

  vec_data_axis.x /= norm;
  vec_data_axis.y /= norm;
  vec_data_axis.z /= norm;


  if( similar_direction_side(vec_data_axis.x, vec_data_axis.y, vec_data_axis.z,
	 datap.nx, datap.ny, datap.nz))
  {

	  similar_side++;

	  //xNǧJɂD

	if (dot_product(vec_data_axis, viewing_direction) > 0.0){
		vec_data_axis.x *= -1.0;
		vec_data_axis.y *= -1.0;
		vec_data_axis.z *= -1.0;
	}	  	  
	  
	  //\߂΁C~ʂƂ̌_oDꂪΉ_ƂȂD

	  foot.x = crossp.x + rad * vec_data_axis.x;
	  foot.y = crossp.y + rad * vec_data_axis.y;
	  foot.z = crossp.z + rad * vec_data_axis.z;

	  dist = sqrt(pow((datap.x - foot.x), 2.0) + pow((datap.y - foot.y), 2.0) + pow((datap.z - foot.z), 2.0));
		/*  Ή_̑gDfobOp  */
		if (disp_errors && disp_errors2)
		{
	/* 
	 *	     fprintf(datafilep, "%f %f %f\n", px, py, pz);
	 *	     fprintf(datafilep, "%f %f %f\n", x1, y1, z1);
	 */
		  printf("side_point_distance\n");
		  fprintf(stderr, "data:(%f, %f, %f) - model:(%f, %f, %f) = %f\n",
			  datap.x, datap.y, datap.z, foot.x, foot.y, foot.z, dist);
		  fprintf(stderr, "[%d] n:(%f, %f, %f)	     n:(%f, %f, %f)\n", 2,
			  pp->nx, pp->ny, pp->nz, vec_data_axis.x, vec_data_axis.y, vec_data_axis.z);
		  getchar();
		}

  }
  else
  {
	  dist = VERY_LARGE;
  }



  return dist;
}

/*
	f[^_~fƓɂ邩ׂ֐
	Ήt̂߂ɎgD
	f[^_PԂ
*/
int check_inside_model(point *pp, cylinder *model)
{
	point datap, center_model_axis, vec_center2data, vec_model_axis;
	double length_proj_vec, cylinder_height_half;

	datap.x = pp->x;
	datap.y = pp->y;
	datap.z = pp->z;
/*
 *	printf("datap = (%f, %f, %f)\n\n",datap.x, datap.y, datap.z);
 */
	//PDf厲̏dSf[^_܂ł̃xNg𓱏o

	center_model_axis.x = (model->disks[0].center.x + model->disks[1].center.x) / 2.0;
	center_model_axis.y = (model->disks[0].center.y + model->disks[1].center.y) / 2.0;
	center_model_axis.z = (model->disks[0].center.z + model->disks[1].center.z) / 2.0;
/*
 *	printf("disks[0].center = (%f, %f, %f)\n",model->disks[0].center.x ,model->disks[0].center.y ,model->disks[0].center.z);
 *	printf("disks[1].center = (%f, %f, %f)\n",model->disks[1].center.x ,model->disks[1].center.y ,model->disks[1].center.z);
 *	printf("center_model_axis = (%f, %f, %f)\n",center_model_axis.x, center_model_axis.y, center_model_axis.z);
 */
	sub_vector(datap, center_model_axis, &vec_center2data);
/*
 *	printf("\n");
 *	printf("vec_center2data = (%f, %f, %f)\n",vec_center2data.x, vec_center2data.y, vec_center2data.z);
 */
	//QDPD̃xNgf厲Ɏˉe

	sub_vector(model->disks[1].center, model->disks[0].center, &vec_model_axis);

	double norm = sqrt(pow(vec_model_axis.x, 2.0) + pow(vec_model_axis.y, 2.0) + pow(vec_model_axis.z, 2.0));

	vec_model_axis.x /= norm;
	vec_model_axis.y /= norm;
	vec_model_axis.z /= norm;
/*
 *	printf("\n");
 *	printf("vec_model_axis = (%f, %f, %f)\n",vec_model_axis.x, vec_model_axis.y, vec_model_axis.z);
 */
	//RD|v| = vEw / |w| @@@v c ˉexNg w c ˉẽxNg
	//
	//@@CˉexNg̒𓾂D
	//    ꂪ~̍̔ȉC~fƓɂƂ݂ȂĂPԂD

	length_proj_vec = dot_product(vec_center2data, vec_model_axis);
/*
 *	printf("\n");
 *	printf("length_proj_vec = %f\n",length_proj_vec);
 */
	cylinder_height_half = distance(model->disks[0].center, model->disks[1].center)/2.0;
/*
 *	printf("\n");
 *	printf("cylinder_height_half = %f\n",cylinder_height_half);
 */
	if(length_proj_vec <= cylinder_height_half)
	{
/*
 *		printf("return 1");
 */
		return 1;
	}
/*	
 *	printf("return 0");
 */
	return 0;
}

/*
  ̕C_Ƃ̋vZD
*/
double polygon_point_distance_for_find_object(double *average, double *normal, point *pp)
{
  double t, nx, ny, nz, px, py, pz, x1, y1, z1, d;
  double dist;

  point foot;

  nx = normal[0];
  ny = normal[1];
  nz = normal[2];

  px = pp->x;
  py = pp->y;
  pz = pp->z;

  /*
    nx x + ny y + nz z + d = 0 Ƃʂ̕ d ߂D
  */
  d = -(  (nx * average[0]) + (ny * average[1]) + (nz * average[2]));

  /*
    ʂ܂ł̕tDʂ̊Oi@jɂƕɂȂD
  */
  t = ((-d) - (nx * px) - (ny * py) - (nz * pz))/(sqr(nx) + sqr(ny) + sqr(nz));

  x1 = px + (t * nx);
  y1 = py + (t * ny);
  z1 = pz + (t * nz);

  foot.x = x1; foot.y = y1; foot.z = z1;


	dist = sqrt(pow(px - x1, 2.0) + pow(py - y1, 2.0) + pow(pz - z1, 2.0));

	/*  Ή_̑gDfobOp  */
	if (disp_errors && disp_errors2)
	{
	/* 
	*	     fprintf(datafilep, "%f %f %f\n", px, py, pz);
	*	     fprintf(datafilep, "%f %f %f\n", x1, y1, z1);
	*/
	}

  return(dist);
}
double disk_point_distance_with_correspondence(disk *diskp, point *pp, point *cp)
{
  double t, nx, ny, nz, px, py, pz, x1, y1, z1, d;
  double dist;

  point foot;

  nx = diskp->nx;
  ny = diskp->ny;
  nz = diskp->nz;

  px = pp->x;
  py = pp->y;
  pz = pp->z;

  /*
    nx x + ny y + nz z + d = 0 Ƃʂ̕ d ߂D
  */
  d = -(  (nx * diskp->center.x) + (ny * diskp->center.y) + (nz * diskp->center.z));

  /*
    ʂ܂ł̕tDʂ̊Oi@jɂƕɂȂD
  */
  t = ((-d) - (nx * px) - (ny * py) - (nz * pz))/(sqr(nx) + sqr(ny) + sqr(nz));

  x1 = px + (t * nx);
  y1 = py + (t * ny);
  z1 = pz + (t * nz);

  foot.x = x1; foot.y = y1; foot.z = z1;

  if(! is_inside_disk(diskp, &foot)){
/* 
 *   if(! is_inside_polygon(polyp, foot)){
 */
    dist = VERY_LARGE;
/* 
 *     dist = distance_to_the_nearest_line(polyp, &foot);
 */
  }
  else{
    dist = sqrt(pow(px - x1, 2.0) + pow(py - y1, 2.0) + pow(pz - z1, 2.0));

    *cp = foot;			/* ݂̐̑Ή_Ƃ */

    /*  Ή_̑gDfobOp  */
    if (disp_errors && disp_errors2)
	{
/* 
 *	     fprintf(datafilep, "%f %f %f\n", px, py, pz);
 *	     fprintf(datafilep, "%f %f %f\n", x1, y1, z1);
 */
		printf("disk_point_distance_with_correspondence\n");
	  fprintf(stderr, "data:(%f, %f, %f) - model:(%f, %f, %f) = %f\n",
		  px, py, pz, x1, y1, z1, dist);
	  fprintf(stderr, "[%d] n:(%f, %f, %f)	     n:(%f, %f, %f)\n", face_num,
		  pp->nx, pp->ny, pp->nz, nx, ny, nz);

	  getchar();
	}
  }

  return(dist);
}
//~̑ʂƓ_Ƃ̋vZD
double side_point_distance_with_correspondence(cylinder *cyp, point *pp, point *cp)
{
  double t, rad;
	  //, nx, ny, nz, px, py, pz, x1, y1, z1, d;
  double dist;

  point sp, ep, crossp, foot, vec_axis, datap, vec_data_axis, viewing_direction;

  viewing_direction.x = viewing_direction.y = 0.0;
  viewing_direction.z = 1.0;

  rad = cyp->disks[0].radius;

  sp = cyp->disks[0].center;
  ep = cyp->disks[1].center;

  vec_axis.x = ep.x - sp.x;
  vec_axis.y = ep.y - sp.y;
  vec_axis.z = ep.z - sp.z;

  double axis_norm = sqrt(pow(vec_axis.x, 2.0) + pow(vec_axis.y, 2.0) + pow(vec_axis.z, 2.0));

  vec_axis.x /= axis_norm;
  vec_axis.y /= axis_norm;
  vec_axis.z /= axis_norm;

  datap.x = pp->x;
  datap.y = pp->y;
  datap.z = pp->z;
  datap.nx = pp->nx;
  datap.ny = pp->ny;
  datap.nz = pp->nz;

  //p[^t𓱏o

  t = dot_product(datap, vec_axis) - dot_product(sp, vec_axis);

  //厲Ƃ̌_xoł

  crossp.x = sp.x + (t * vec_axis.x);
  crossp.y = sp.y + (t * vec_axis.y);
  crossp.z = sp.z + (t * vec_axis.z);

  //厲Ƃ̌_f[^_ւ̃xNgƁCf[^_̖@Ƃ̂Ȃp

  vec_data_axis.x = datap.x - crossp.x;
  vec_data_axis.y = datap.y - crossp.y;
  vec_data_axis.z = datap.z - crossp.z;

  double norm = sqrt(pow(vec_data_axis.x, 2.0) + pow(vec_data_axis.y, 2.0) + pow(vec_data_axis.z, 2.0));

  vec_data_axis.x /= norm;
  vec_data_axis.y /= norm;
  vec_data_axis.z /= norm;

  if( similar_direction_side(vec_data_axis.x, vec_data_axis.y, vec_data_axis.z,
	 datap.nx, datap.ny, datap.nz))
  {

	  similar_side_cp++;

	  //\߂΁C~ʂƂ̌_oDꂪΉ_ƂȂD

	//xNǧJɂD
	if (dot_product(vec_data_axis, viewing_direction) > 0.0){
		vec_data_axis.x *= -1.0;
		vec_data_axis.y *= -1.0;
		vec_data_axis.z *= -1.0;
	}

	  foot.x = crossp.x + rad * vec_data_axis.x;
	  foot.y = crossp.y + rad * vec_data_axis.y;
	  foot.z = crossp.z + rad * vec_data_axis.z;

	  //Ή_Ƃ
	  *cp = foot;

	  dist = sqrt(pow((datap.x - foot.x), 2.0) + pow((datap.y - foot.y), 2.0) + pow((datap.z - foot.z), 2.0));


	  /*  Ή_̑gDfobOp  */
		if (disp_errors && disp_errors2)
		{
	/* 
	 *	     fprintf(datafilep, "%f %f %f\n", px, py, pz);
	 *	     fprintf(datafilep, "%f %f %f\n", x1, y1, z1);
	 */
		  printf("side_point_distance_with_correspondence\n");
		  fprintf(stderr, "data:(%f, %f, %f) - model:(%f, %f, %f) = %f\n",
			  datap.x, datap.y, datap.z, foot.x, foot.y, foot.z, dist);
		  fprintf(stderr, "[%d] n:(%f, %f, %f)	     n:(%f, %f, %f)\n", 2,
			  pp->nx, pp->ny, pp->nz, vec_data_axis.x, vec_data_axis.y, vec_data_axis.z);
		  getchar();
		}

  }
  else
  {
	  dist = VERY_LARGE;
  }

  return dist;
}
/*
  p`Ɠ_Ƃ̋vZD
  _瑽p`ɉ낵̑ʏɂȂƂɂ́C傫lԂD
*/
double polygon_point_distance(polygon *polyp, point *pp)
{
  double t, nx, ny, nz, px, py, pz, x1, y1, z1, d;
  double dist;

  point foot;

  nx = polyp->nx;
  ny = polyp->ny;
  nz = polyp->nz;

  px = pp->x;
  py = pp->y;
  pz = pp->z;

  /*
    nx x + ny y + nz z + d = 0 Ƃʂ̕ d ߂D
  */
  d = -(  (nx * polyp->lines[0].sp.x) + (ny * polyp->lines[0].sp.y) + (nz * polyp->lines[0].sp.z));

  /*
    ʂ܂ł̕tDʂ̊Oi@jɂƕɂȂD
  */
  t = ((-d) - (nx * px) - (ny * py) - (nz * pz))/(sqr(nx) + sqr(ny) + sqr(nz));

  x1 = px + (t * nx);
  y1 = py + (t * ny);
  z1 = pz + (t * nz);

  foot.x = x1; foot.y = y1; foot.z = z1;

  if(! is_inside_polygon2(polyp, &foot)){
/* 
 *   if(! is_inside_polygon(polyp, foot)){
 */
    dist = VERY_LARGE;
/* 
 *     dist = distance_to_the_nearest_line(polyp, &foot);
 */
  }
  else{
    dist = sqrt(pow(px - x1, 2.0) + pow(py - y1, 2.0) + pow(pz - z1, 2.0));

    /*  Ή_̑gDfobOp  */
    if (disp_errors && disp_errors2)
	{
/* 
 *	     fprintf(datafilep, "%f %f %f\n", px, py, pz);
 *	     fprintf(datafilep, "%f %f %f\n", x1, y1, z1);
 */
	  fprintf(stderr, "data:(%f, %f, %f) - model:(%f, %f, %f) = %f\n",
		  px, py, pz, x1, y1, z1, dist);
	  fprintf(stderr, "[%d] n:(%f, %f, %f)	     n:(%f, %f, %f)\n", face_num,
		  pp->nx, pp->ny, pp->nz, nx, ny, nz);
	}
  }

  return(dist);
}
/*
  p`Ɠ_Ƃ̋vZD
  _瑽p`ɉ낵̑ʏɂȂƂɂ́C傫lԂD
  ʏɂƂɂ͋Ɛ̑ԂD
*/
double polygon_point_distance_with_correspondence(polygon *polyp, point *pp, point *cp)
{
  double t, nx, ny, nz, px, py, pz, x1, y1, z1, d;
  double dist;

  point foot;

  nx = polyp->nx;
  ny = polyp->ny;
  nz = polyp->nz;

  px = pp->x;
  py = pp->y;
  pz = pp->z;

  /*
    nx x + ny y + nz z + d = 0 Ƃʂ̕ d ߂D
  */
  d = -(  (nx * polyp->lines[0].sp.x) + (ny * polyp->lines[0].sp.y) + (nz * polyp->lines[0].sp.z));

  /*
    ʂ܂ł̕tDʂ̊Oi@jɂƕɂȂD
  */
  t = ((-d) - (nx * px) - (ny * py) - (nz * pz))/(sqr(nx) + sqr(ny) + sqr(nz));

  x1 = px + (t * nx);
  y1 = py + (t * ny);
  z1 = pz + (t * nz);

  foot.x = x1; foot.y = y1; foot.z = z1;

  if(! is_inside_polygon2(polyp, &foot)){
/* 
 *   if(! is_inside_polygon(polyp, foot)){
 */
    dist = VERY_LARGE;
/* 
 *     dist = distance_to_the_nearest_line(polyp, &foot);
 */
  }
  else{
    dist = sqrt(pow(px - x1, 2.0) + pow(py - y1, 2.0) + pow(pz - z1, 2.0));

    *cp = foot;			/* ݂̐̑Ή_Ƃ */

    /*  Ή_̑gDfobOp  */
    if (disp_errors && disp_errors2)
	{
/* 
 *	     fprintf(datafilep, "%f %f %f\n", px, py, pz);
 *	     fprintf(datafilep, "%f %f %f\n", x1, y1, z1);
 */
	  fprintf(stderr, "data:(%f, %f, %f) - model:(%f, %f, %f) = %f\n",
		  px, py, pz, x1, y1, z1, dist);
	  fprintf(stderr, "[%d] n:(%f, %f, %f)	     n:(%f, %f, %f)\n", face_num,
		  pp->nx, pp->ny, pp->nz, nx, ny, nz);
	}
  }

  return(dist);
}

/*
 *  2̖@߂ǂD
 *  @xNgŕ\Cdot_product֐gׂ
 */
#define NORMAL_THRESH deg_to_rad(30.0)
int similar_direction(double nx1, double ny1, double nz1,
		      double nx2, double ny2, double nz2){

  double dotp;

  dotp = nx1*nx2 + ny1*ny2 + nz1*nz2;
  if (dotp > cos(NORMAL_THRESH)) return 1;
  else return 0;
}

//v
#define NORMAL_THRESH2 deg_to_rad(15.0)
int similar_direction2(double nx1, double ny1, double nz1,
		      double nx2, double ny2, double nz2){

  double dotp;

  dotp = nx1*nx2 + ny1*ny2 + nz1*nz2;
  if (dotp > cos(NORMAL_THRESH)) return 1;
  else return 0;
}
//ʂ̓_Qop
//#define NORMAL_THRESH3 deg_to_rad(35.0)
#define NORMAL_THRESH3 deg_to_rad(30.0)
int similar_direction3(double nx1, double ny1, double nz1,
		      double nx2, double ny2, double nz2){

  double dotp;

  dotp = nx1*nx2 + ny1*ny2 + nz1*nz2;
  if (dotp > cos(NORMAL_THRESH3)) return 1;
  else return 0;
}

//ʂ̑Ήtp
//ς̐ΒlƂD
#define NORMAL_THRESH_SIDE deg_to_rad(20.0)
//#define NORMAL_THRESH_SIDE deg_to_rad(25.0)
int similar_direction_side(double nx1, double ny1,double nz1,
						double nx2, double ny2, double nz2)
{
	double dotp;
	//1`-1ŕω̂ŁCΒlC1`0ɂ
	dotp = nx1*nx2 + ny1*ny2 + nz1*nz2;	

	//ΒlƂƋtƑΉĂ܂̂ł
//	if (fabs(dotp) > cos(NORMAL_THRESH_SIDE)) return 1;
	if (dotp > cos(NORMAL_THRESH_SIDE)) return 1;
	else return 0;
}

//--------------------------------------------------------------------------------------//
/*  calcNormal: ʃpb`璆_̖@PCAŋ߂  */
// CvArr  CvPoint3D32f ܂H

// tZbg̃f[^ȂΖ@(0, 0, 0)Ƃ @̊֐ĂԑOɃ`FbN

#define DATA_DIM	3
#define SurfaceNormalWindowSize	7
#define HalfSurfaceNormalWindowSize	3
#define SurfaceThresh	0.1

double myDotProduct(CvPoint3D32f v1, CvPoint3D32f v2){
	return(v1.x * v2.x + v1.y * v2.y + v1.z * v2.z);
}
//--------------------------------------------------------------------------------------//
void calcNormal(CvPoint3D32f points[][SurfaceNormalWindowSize], CvPoint3D32f *normal)
{
	int i, j;
	int valid_point_num = 0;	// L_f[^

	CvMat	*source,			// ͓_f[^
			*average,			// σxNgidSj
			*eigenValues,		// ŗLl
			*eigenVectors;		// ŗLxNg

	CvPoint3D32f surface_normal, viewing_direction;

	// ̐ݒ
	viewing_direction.x = viewing_direction.y = 0.0;
	viewing_direction.z = 1.0;

	// Ɨpf[^̊蓖
	source = cvCreateMat(sqr(SurfaceNormalWindowSize), DATA_DIM, CV_32F);	// Rawf[^
	average = cvCreateMat(1, DATA_DIM, CV_32F);								// σxNg
	eigenValues = cvCreateMat(1, PCA_DIM, CV_32F);
	eigenVectors = cvCreateMat(PCA_DIM, DATA_DIM, CV_32F);

	// f[^̃Zbg
	valid_point_num = 0;
	for (i = 0; i < SurfaceNormalWindowSize; i++){
		for (j = 0; j < SurfaceNormalWindowSize; j++){
			cvmSet(source, valid_point_num, 0, points[i][j].x);
			cvmSet(source, valid_point_num, 1, points[i][j].y);
			cvmSet(source, valid_point_num, 2, points[i][j].z);

			valid_point_num++;
		}
	}

	// 听́iPCAj
	cvCalcPCA(source, average, eigenValues, eigenVectors, CV_PCA_DATA_AS_ROW);

	// 3ŗLxNgʂ̖@ɑΉ
	// ʂǂ͑3ŗLl̑傫Ŕ
	if (cvmGet(eigenValues, 0, 2) > SurfaceThresh){
		// @܂ȂƂāC(0, 0, 0) Ƃ
		normal->x = normal->y = normal->z = 0.0;

		return;
	}

	surface_normal.x = cvmGet(eigenVectors, 2, 0);
	surface_normal.y = cvmGet(eigenVectors, 2, 1);
	surface_normal.z = cvmGet(eigenVectors, 2, 2);

	// 3ŗLxNg͗iȂƓjĂ邩Ȃ̂
	// ̏ꍇ̓J悤ɂ
	if (myDotProduct(surface_normal, viewing_direction) > 0.0){
		surface_normal.x *= -1.0;
		surface_normal.y *= -1.0;
		surface_normal.z *= -1.0;
	}

	*normal = surface_normal;

	// 
	cvReleaseMat(&source);
	cvReleaseMat(&average);
	cvReleaseMat(&eigenValues);
	cvReleaseMat(&eigenVectors);
}

void calcSurfacePCA(CvPoint3D32f *averagep, CvPoint3D32f *vec1, CvPoint3D32f *vec2, CvPoint3D32f *vec3, 
					double *xval, double *yval, double *zval, int pointnum)
{
	//ʂ̕Zo铮ǉ(11/09/19)

	CvMat	*source,			// ͓_f[^
			*average,			// σxNgidSj
			*eigenValues,		// ŗLl
			*eigenVectors;		// ŗLxNg

	CvPoint3D32f surface_normal, viewing_direction;
	CvPoint3D32f tempAverage, tempVec1, tempVec2, tempVec3;

	// ̐ݒ
	viewing_direction.x = viewing_direction.y = 0.0;
	viewing_direction.z = 1.0;

	// Ɨpf[^̊蓖
	//cvCreateMat̎g: (s, , CV_32F)
	source  = cvCreateMat(pointnum,DATA_DIM,CV_32F);  
	average = cvCreateMat(1, DATA_DIM, CV_32F);								// σxNg
	eigenValues = cvCreateMat(1, PCA_DIM, CV_32F);
	eigenVectors = cvCreateMat(PCA_DIM, DATA_DIM, CV_32F);

	// wif[^̃Zbg
	for (int i = 0; i < pointnum; i++){
		cvmSet(source, i, 0, xval[i]);
		cvmSet(source, i, 1, yval[i]);
		cvmSet(source, i, 2, zval[i]);
	}

	// 听́iPCAj
	// 3ŗLxNgʂ̖@ɑΉ
	cvCalcPCA(source, average, eigenValues, eigenVectors, CV_PCA_DATA_AS_ROW);


	tempAverage.x = cvmGet(average, 0, 0);
	tempAverage.y = cvmGet(average, 0, 1);
	tempAverage.z = cvmGet(average, 0, 2);

	tempVec1.x = cvmGet(eigenVectors, 0, 0);
	tempVec1.y = cvmGet(eigenVectors, 0, 1);
	tempVec1.z = cvmGet(eigenVectors, 0, 2);

	tempVec2.x = cvmGet(eigenVectors, 1, 0);
	tempVec2.y = cvmGet(eigenVectors, 1, 1);
	tempVec2.z = cvmGet(eigenVectors, 1, 2);

	tempVec3.x = cvmGet(eigenVectors, 2, 0);
	tempVec3.y = cvmGet(eigenVectors, 2, 1);
	tempVec3.z = cvmGet(eigenVectors, 2, 2);

	printf("vec3 before \n");
	printf("vec3(%3.3f, %3.3f, %3.3f)\n",tempVec3.x, tempVec3.y, tempVec3.z);

	//@ꍇCɏC

	// 3ŗLxNg͗iȂƓjĂ邩Ȃ̂
	// ̏ꍇ̓J悤ɂ
	if (myDotProduct(tempVec3, viewing_direction) > 0.0){
		tempVec3.x *= -1.0;
		tempVec3.y *= -1.0;
		tempVec3.z *= -1.0;
	}

	printf("vec3 after \n");
	printf("vec3(%3.3f, %3.3f, %3.3f)\n",tempVec3.x, tempVec3.y, tempVec3.z);

	*averagep = tempAverage;
	*vec1 = tempVec1;
	*vec2 = tempVec2;
	*vec3 = tempVec3;

}

/*
 * ̕ʂ̕𓱏o
 */
void calcTablePlane(double *xval, double *yval, double *zval)
{
	CvMat	*source,			// ͓_f[^
			*average,			// σxNgidSj
			*eigenValues,		// ŗLl
			*eigenVectors;		// ŗLxNg

	CvPoint3D32f surface_normal, viewing_direction;

	// ̐ݒ
	viewing_direction.x = viewing_direction.y = 0.0;
	viewing_direction.z = 1.0;

	int plane_count = 0;

	//for debug 1018
//	printf("DistanceThresh : %.3f\n",DistanceThresh);

	//߂Ő^񒆂̉fg

	plane_count = 88*72;

	//for debug 1017
	printf("plane_count:%d\n",plane_count);

	// Ɨpf[^̊蓖
	//cvCreateMat̎g: (s, , CV_32F)
	source  = cvCreateMat(plane_count, DATA_DIM, CV_32F);
	average = cvCreateMat(1, DATA_DIM, CV_32F);								// σxNg
	eigenValues = cvCreateMat(1, PCA_DIM, CV_32F);
	eigenVectors = cvCreateMat(PCA_DIM, DATA_DIM, CV_32F);

	// wif[^̃Zbg
	//߂Ő^񒆂̉fg

	int ind, i = 0;

	std::ofstream ofsBackg, ofsBackgAll;
	ofsBackg.open("backbround_center.txt",std::ios::out);
	ofsBackgAll.open("backbround.txt",std::ios::out);

	for(int h = 0; h < (HEIGHT_VALUE); h++)
	{
		for(int w = 0; w < (WIDTH_VALUE); w++)
		{
			ind = h * WIDTH_VALUE + w;

			ofsBackgAll << xval[ind] << ' ' << yval[ind] << ' ' << zval[ind] << std::endl;

		}
	}
	ofsBackgAll.close();

	for(int h = 35; h < (HEIGHT_VALUE - 37); h++)
	{
		for(int w = 43; w < (WIDTH_VALUE - 45); w++)
		{
			ind = h * WIDTH_VALUE + w;

			cvmSet(source, i, 0, xval[ind]);
			cvmSet(source, i, 1, yval[ind]);
			cvmSet(source, i, 2, zval[ind]);

			ofsBackg << xval[ind] << ' ' << yval[ind] << ' ' << zval[ind] << std::endl;

			i += 1;

		}
	}

	ofsBackg.close();

	printf("plane_count:%d\n",plane_count);

	// 听́iPCAj
	// 3ŗLxNgʂ̖@ɑΉ
	cvCalcPCA(source, average, eigenValues, eigenVectors, CV_PCA_DATA_AS_ROW);

	//@ꍇCɏC

	surface_normal.x = cvmGet(eigenVectors, 2, 0);
	surface_normal.y = cvmGet(eigenVectors, 2, 1);
	surface_normal.z = cvmGet(eigenVectors, 2, 2);

	// 3ŗLxNg͗iȂƓjĂ邩Ȃ̂
	// ̏ꍇ̓J悤ɂ
	if (myDotProduct(surface_normal, viewing_direction) > 0.0){
		surface_normal.x *= -1.0;
		surface_normal.y *= -1.0;
		surface_normal.z *= -1.0;
	}

	printf("average(%3.3f, %3.3f, %3.3f)\n"
		,cvmGet(average,0,0), cvmGet(average,0,1), cvmGet(average,0,2));
	printf("PCA3(%3.3f, %3.3f, %3.3f)\n"
		,surface_normal.x, surface_normal.y, surface_normal.z);

	std::ofstream ofs;	
	ofs.open("planeparameter.txt",std::ios::out);

	ofs << cvmGet(average,0,0) << ',' << cvmGet(average,0,1) << ',' << cvmGet(average,0,2) << std::endl;
	ofs << surface_normal.x << ',' << surface_normal.y << ',' << surface_normal.z << std::endl;

	ofs.close();

	//̊J
	cvReleaseMat(&source);
	cvReleaseMat(&average);
	cvReleaseMat(&eigenValues);
	cvReleaseMat(&eigenVectors);
}

// convert_point_data: f[^_f[^ɕϊD
//    f[^ x, y, z`ŁCforexval, foreyval, forezval ɕۑĂD
//    _f[^͖@f[^Cpoints[], point_num ̃Xg`

void convert_point_data(int *point_num, point *points)
{
	int w, h, ind;
	int ww, hh, indd;
	int p_num = 0;

	CvPoint3D32f normal;
	CvPoint3D32f surface_region[SurfaceNormalWindowSize][SurfaceNormalWindowSize];

	int appear_background, dont_calc_normal;
	appear_background = dont_calc_normal = 0;

	for (h = 0; h < data_height; h++){
		for (w = 0; w < data_width; w++){
			ind = h * data_width + w;
//			printf("h %d w %d\n",h,w);
//			printf("ind %d\n",ind);
			if (forezval[ind] != BACKGROUND_DEPTH){ // Ώە̂̃f[^Ȃ
				// ̓_𒆐SƂǏ̈ɖʃtBbeBOsĖ@ZoD
				// Ǐ̈ɔwi܂܂ĂCΏۃsNZ͂D
				for (hh = -HalfSurfaceNormalWindowSize; hh <= HalfSurfaceNormalWindowSize; hh++){
					for (ww = -HalfSurfaceNormalWindowSize; ww <= HalfSurfaceNormalWindowSize; ww++){
						indd = ind + hh * data_width + ww;
//						printf("hh %d ww %d\n",hh,ww);
//						printf("indd %d\n",indd);

						if (forezval[indd] == BACKGROUND_DEPTH){ // Ǐ̈ɔwif[^
							appear_background++;
							goto SKIP;
						}

						surface_region[hh+HalfSurfaceNormalWindowSize][ww+HalfSurfaceNormalWindowSize].x = forexval[indd]; 
						surface_region[hh+HalfSurfaceNormalWindowSize][ww+HalfSurfaceNormalWindowSize].y = foreyval[indd]; 
						surface_region[hh+HalfSurfaceNormalWindowSize][ww+HalfSurfaceNormalWindowSize].z = forezval[indd]; 
					}
				}

				calcNormal(surface_region, &normal);
				if ((normal.x == 0.0) && (normal.y == 0.0) && (normal.z == 0.0)){ // @܂ȂiȂǁj
					dont_calc_normal++;
				}
				else{
					points[p_num].x = forexval[ind];
					points[p_num].y = foreyval[ind];
					points[p_num].z = forezval[ind];
					points[p_num].nx = normal.x;
					points[p_num].ny = normal.y;
					points[p_num].nz = normal.z;
					p_num++;
				}
			}
			SKIP:;
		}
	}
	//appear_backgroundœ_菜ĂD
	printf("appear_background:%d\n",appear_background);
	printf("dont_calc_normal:%d\n",dont_calc_normal);
	*point_num = p_num;
}
