#pragma once


#include "cvmttype.h"
#include "cvmtvector.h"
#include <iostream>


namespace opencv
{
namespace cvmath
{
	
	/**
		sNX
	*/
	template <int CV_TYPE>
	class Matrix{
	public:
		CvMat *e;

	public:	// Constructors & destructor
		Matrix(){	e = NULL;	}
		Matrix(int r, int c)
		{	e = cvCreateMat(r, c, CV_TYPE);	}
		
		~Matrix()
		{
			cvReleaseMat(&e);
			e = NULL;
		}

		void resize(int r, int c)
		{
			if(e != NULL)
				cvReleaseMat(&e);
			e = cvCreateMat(r, c, CV_TYPE);
		}


	public:

		void set(int i, int j, double val)
		{	cvmSet(e, i, j, val);	}

		typename CVElem::CvElemTypeTraits<CV_TYPE>::ElemType get(int i, int j) const
		{	return (CVElem::CvElemTypeTraits<CV_TYPE>::ElemType) cvmGet(e, i, j);	}
		
		void set(const double *M)
		{
			for(int r = 0; r < e->rows; r ++)
				for(int c = 0; c < e->cols; c ++)
					set(r, c, *(M ++));
		}

		void set(const Matrix &M)
		{	cvCopy(M.e, e);		}

	public:
		// Set matrix
		void set(int i, int j, const Matrix &M)
		{
			for(int y = 0; y < M.e->rows; y ++)
				for(int x = 0; x < M.e->cols; x ++)
					set(i + y, j + x, M.get(y, x));
		}

		// Set transposed matrix
		void setT(int i, int j, const Matrix &M)
		{
			for(int y = 0; y < M.e->rows; y ++)
				for(int x = 0; x < M.e->cols; x ++)
					set(i + x, j + y, M.get(y, x));
		}

		// Set column vector
		void set(int i, int j, const Vector<CV_TYPE> &v)
		{
			for(int y = 0; y < v.e->rows; y ++)
				set(i + y, j, v.get(y));
		}

		// Set row vector
		void setT(int i, int j, const Vector<CV_TYPE> &v)
		{
			for(int x = 0; x < v.e->rows; x ++)
				set(i, j + x, v.get(x));
		}

		// Get sub matrix
		void get(int i, int j, Matrix &M) const
		{
			for(int y = 0; y < M.e->rows; y ++)
				for(int x = 0; x < M.e->cols; x ++)
					M.set(y, x, get(i + y, j + x));
		}

		// Get sub matrix from transposed matrix
		void getT(int i, int j, Matrix &M) const
		{
			for(int y = 0; y < M.e->rows; y ++)
				for(int x = 0; x < M.e->cols; x ++)
					M.set(y, x, get(i + x, j + y));
		}

		// Get sub vector
		void get(int i, int j, Vector<CV_TYPE> &v) const
		{
			for(int vi = 0; vi < v.e->rows; ++ vi)
				v.set(vi, get(i + vi, j));
		}

		// Get sub vector from transposed matrix
		void getT(int i, int j, Vector<CV_TYPE> &v) const
		{
			for(int vi = 0; vi < v.e->rows; ++ vi)
				v.set(vi, get(i, j + vi));
		}

	public:
		// Zero clear
		void setZero(){	cvZero(e);	}

		// Set all elements
		void setAll(double val)
		{	cvSet(e, cvScalarAll(val));	}

		// Identity matrix
		void setI(double val = 1.0)
		{	cvSetIdentity(e, cvScalarAll(val));	}

		// Transpose
		void trans(Matrix<CV_TYPE> &Mt) const
		{	cvmTranspose(e, Mt.e);	}

		// Invert matrix
		void invert(Matrix<CV_TYPE> &Minv) const
		{	cvmInvert(e, Minv.e);	}

		// Invert matrix (optimized for 2x2)
		void invert22(Matrix<CV_TYPE> &Minv) const
		{
			const double a = (*this)(0, 0), b = (*this)(0, 1), c = (*this)(1, 0), d = (*this)(1, 1), ad_minus_bc = a*d - b*c;
			Minv.set(0, 0, d / ad_minus_bc);	Minv.set(0, 1, -b / ad_minus_bc);
			Minv.set(1, 0, -c / ad_minus_bc);	Minv.set(1, 1, a / ad_minus_bc);	
		}

		// Trace
		double trace() const
		{
			return cvmTrace(e);
		}

	public:
		/* 2s񎮂̒l */
		typename CVElem::CvElemTypeTraits<CV_TYPE>::ElemType det22() const
		{
			return get(0, 0) * get(0, 1) - get(1, 0) * get(0, 1);	
		}

		/**	s̃XJ[{	*/
		void xS(double s)
		{
			for(int r = 0; r < e->rows; r ++)
				for(int c = 0; c < e->cols; c ++)
					set(r, c, get(r, c) * s);
		}

		/**	s̃XJ[{	*/
		void SxM(double s, const Matrix &M)
		{
			cvCopy(M.e, e);
			xS(s);
		}

		/**	s alpha*AB + beta*C ̌vZ	*/
		void MxM(const Matrix &A, const Matrix &B, const double alpha = 1.0, const double beta = 0.0)
		{
			cvGEMM(A.e, B.e, alpha, e, beta, e);
		}

		/**	s alpha*A^tB + beta*C ̌vZ	*/
		void MtxM(const Matrix &A, const Matrix &B, const double alpha = 1.0, const double beta = 0.0)
		{	cvGEMM(A.e, B.e, alpha, e, beta, e, CV_GEMM_A_T);	}

		/**	s alpha*AB^t + beta*C ̌vZ	*/
		void MxMt(const Matrix &A, const Matrix &B, const double alpha = 1.0, const double beta = 0.0)
		{	cvGEMM(A.e, B.e, alpha, e, beta, e, CV_GEMM_B_T);	}

		/**	s alpha*A^tB^t + beta*C ̌vZ	*/
		void MtxMt(const Matrix &A, const Matrix &B, const double alpha = 1.0, const double beta = 0.0)
		{	cvGEMM(A.e, B.e, alpha, e, beta, e, CV_GEMM_A_T + CV_GEMM_B_T);	}

		/**	s AtBA ̌vZ	*/
		void MtxMxM(const Matrix &A, const Matrix &B, Matrix &AtB)
		{
			AtB.MtxM(A, B);
			MxM(AtB, A);
		}

		/**	s ABAt ̌vZ	*/
		void MxMxMt(const Matrix &A, const Matrix &B, Matrix &AB)
		{
			AB.MxM(A, B);
			MxMt(AB, A);
		}

		/**	saA+B̌vZ	*/
		void MplusM(const Matrix &A, const Matrix &B){	cvmAdd(A.e, B.e, e);	}

		/**	sA-B̌vZ	*/
		void MminusM(const Matrix &A, const Matrix &B){	cvmSub(A.e, B.e, e);	}

	public:
		typedef typename CVElem::CvElemTypeTraits<CV_TYPE>::ElemType ElemType;
		typename ElemType &operator ()(int r, int c)	{	return ((ElemType *) e->data.ptr)[r*e->cols + c];	}
		typename const ElemType &operator ()(int r, int c)	const {	return ((ElemType *) e->data.ptr)[r*e->cols + c];	}

		operator CvMat *(){ return e;	}
		operator const CvMat *() const{	return e;	}

	public:
		void print() const
		{
			for(int i = 0; i < e->rows; i ++){
				std::cout << "|";
				for(int j = 0; j < e->cols; j ++)
					std::cout << " " << get(i, j) << " ";
				std::cout << "|" << std::endl;
			}
		}


	};

}	// namespace Math
}