#pragma once


#include "cvmttype.h"


namespace opencv
{
namespace cvmath
{


	/**
		lNX
	*/
	template <int CV_TYPE>
	class Quaternion : public Vector<CV_TYPE>
	{
	public:	// Constructors & destructor
		Quaternion() : Vector(4){}

		Quaternion(const Vector &axis) : Vector(4)
		{	fromAxis(axis);	}
		
		~Quaternion(){}

	public:
		void setR(double r){	set(0, r);	}
		void setX(double x){	set(1, x);	}
		void setY(double y){	set(2, y);	}
		void setZ(double z){	set(3, z);	}
		void setI(){	setR(1.0);	setX(0.0);	setY(0.0);	setZ(0.0);	}

		double getR() const{	return get(0);	}
		double getX() const{	return get(1);	}
		double getY() const{	return get(2);	}
		double getZ() const{	return get(3);	}

	public:
		void fromAxis(const Vector &axis)
		{
			const double angle = axis.norm();
			if(angle > 0.0){
				const double COS = cos(angle / 2.0);
				const double SIN = sin(angle / 2.0) / angle;

				setR(COS);	setX(SIN * axis.get(0));	setY(SIN * axis.get(1));	setZ(SIN * axis.get(2));
			}
			else{	setI();	}
		}

		void conjugate(Quaternion &Qc)
		{
			Qc.setR( getR());
			Qc.setX(-getX());
			Qc.setY(-getY());
			Qc.setZ(-getZ());
		}
		
		void invert()
		{
			const double r = getR(), x = getX(), y = getY(), z = getZ();
			const double coef = 1.0 / (r*r + x*x + y*y + z*z);
			setR( r*coef);
			setX(-x*coef);
			setY(-y*coef);
			setZ(-z*coef);
		}

		void invert(Quaternion &Qinv) const
		{
			const double r = getR(), x = getX(), y = getY(), z = getZ();
			const double coef = 1.0 / (r*r + x*x + y*y + z*z);
			Qinv.setR( r*coef);
			Qinv.setX(-x*coef);
			Qinv.setY(-y*coef);
			Qinv.setZ(-z*coef);
		}

		void toR(Matrix<CV_TYPE> &R) const
		{
			const double r = getR(), x = getX(), y = getY(), z = getZ();

			// Ίpvf
			const double r2 = r * r, x2 = x * x, y2 = y * y, z2 = z * z;
			R.set(0, 0, r2 + x2 - y2 - z2);
			R.set(1, 1, r2 - x2 + y2 - z2);
			R.set(2, 2, r2 - x2 - y2 + z2);

			// Ίpvf
			const double xy = x*y, yz = y*z, zx = z*x, rx = r*x, ry = r*y, rz = r*z;
																	R.set(0, 1, 2 * (xy - rz));	R.set(0, 2, 2 * (zx + ry));
			R.set(1, 0, 2 * (xy + rz));															R.set(1, 2, 2 * (yz - rx));
			R.set(2, 0, 2 * (zx - ry));	R.set(2, 1, 2 * (yz + rx));
		}

		void QxQ(const Quaternion &A, const Quaternion &B)
		{
			const double ar = A.getR(), ax = A.getX(), ay = A.getY(), az = A.getZ();
			const double br = B.getR(), bx = B.getX(), by = B.getY(), bz = B.getZ();
			
			setR(ar*br - ax*bx - ay*by - az*bz);
			setX(ar*bx + ax*br + ay*bz - az*by);
			setY(ar*by - ax*bz + ay*br + az*bx);
			setZ(ar*bz + ax*by - ay*bx + az*br);
		}

	};


}
}