#pragma once


#include "cvmttype.h"
#include <iostream>


namespace opencv
{
namespace cvmath
{

	template <int CV_TYPE>
	class Matrix;	// s錾


	/**
		xNgNX
	*/
	template <int CV_TYPE>
	class Vector{
	public:
		CvMat *e;

	public:	// Constructors & destructor
		Vector(){	e = NULL;	}

		Vector(int dim)
		{	e = cvCreateMat(dim, 1, CV_TYPE);	}

		virtual ~Vector()
		{
			cvReleaseMat(&e);
			e = NULL;
		}
	
		void resize(int dim)
		{
			if(e != NULL)
				cvReleaseMat(&e);
			e = cvCreateMat(dim, 1, CV_TYPE);
		}

		void setZero()
		{	cvZero(e);	}

	public:	/* v~eBudl */
		/**	iԖڂ̗vfvalݒ肷	*/
		void set(int i, double val)
		{	cvmSet(e, i, 0, val);	}

		/**	iԖڂ̗vf̎擾	*/
		typedef typename CVElem::CvElemTypeTraits<CV_TYPE>::ElemType elem_t;
		elem_t get(int i) const
		{	return elem_t(cvmGet(e, i, 0));	}

		/**	iԖڂ̗vfɔzvdimvfݒ肷	*/
		void set(int i, const double *v, int dim)
		{	for(int e_i = i, v_i = 0; v_i < dim; e_i ++, v_i ++)	set(e_i, v[v_i]);	}

	public:
		/**	xNgvݒ肷	*/
		void set(const Vector &v){	cvCopy(v.e, e);	}

		/**	iԖڂ̗vfɃxNgvݒ肷	*/
		void set(const int i, const Vector &v)
		{
			for(int e_i = i, v_i = 0; v_i < v.e->rows; e_i ++, v_i ++)
				set(e_i, v.get(v_i));
		}

		/**	iԖڂ̗vfɁuxNgvviԖڗvfvnvݒ	*/
		void set(int i, const Vector &v, int vi, int vn)
		{
			for(int e_i = i, v_i = vi; v_i < vn; e_i ++, v_i ++)
				set(e_i, v.get(v_i));
		}

	public:
		/**	xNg2m̌vZ	*/
		double norm() const
		{
			double norm_sq = 0;
			for(int i = 0; i < e->rows; i ++)
				norm_sq += get(i) * get(i);
			return sqrt(norm_sq);
		}

		/**	xNgK	*/
		void normalize()
		{
			const double scaler = 1.0 / norm();
			xS(scaler);
		}

		/**	xNg̃XJ[{	*/
		void xS(double s){
			for(int i = 0; i < e->rows; i ++)
				set(i, get(i) * s);
		}

		/**	xNg̃XJ[{	*/
		void SxV(double s, const Vector &v)
		{
			cvCopy(v.e, e);
			xS(s);
		}

		/**	xNgax+y̌vZ	*/
		void VplusV(const Vector &a, const Vector &b)
		{	cvmAdd(a.e, b.e, e);	}

		/**	xNgx-y̌vZ	*/
		void VminusV(const Vector &a, const Vector &b)
		{	cvmSub(a.e, b.e, e);	}

		/**	xNg	*/
		double dotV(const Vector &a)
		{	return cvmDotProduct(a.e, e);	}

		/**	xNg alpha*A*a + beta*b ̌vZ	*/
		void MxV(const Matrix<CV_TYPE> &A, const Vector &a, const double alpha = 1.0, const double beta = 0.0)
		{	cvGEMM(A.e, a.e, alpha, e, beta, e);	}

		/**	xNgO(3)	*/
		void VcrossV3(const Vector &a, const Vector &b)
		{
			const elem_t a0 = a.get(0), a1 = a.get(1), a2 = a.get(2);
			const elem_t b0 = b.get(0), b1 = b.get(1), b2 = b.get(2);
			set(0, a1*b2 - b1*a2);
			set(1, a2*b0 - b2*a0);
			set(2, a0*b1 - b0*a1);
		}

	public:
		typedef typename CVElem::CvElemTypeTraits<CV_TYPE>::ElemType ElemType;
		typename ElemType &operator ()(int i){	return ((ElemType *) e->data.ptr)[i];	}
		const typename ElemType &operator ()(int i)	const	{	return ((ElemType *) e->data.ptr)[i];	}


		operator CvMat *(){	return e;	}
		operator const CvMat *() const{	return e;	}

	public:
		void print() const{
			std::cout << "[ ";
			for(int i = 0; i < e->rows; i ++)
				std::cout << " " << get(i) << " ";
			std::cout << "]" << std::endl;
		}

	};

}	// namespace Math
}