#include "../cvmath/cvmath.h"
#include "../cvutil/cvutil.h"

#include <iostream>
#include <sstream>
#include <fstream>

#include "../../icp3d.h"

#include <math.h>

using namespace std;

static const char *IMG_PATH = "background.png";
static const char *XML_PATH = "camera.xml";

bool isParameterLoaded = false;
opencv::cvmath::Matrix<CV_64F> intrinsic, distortion;
opencv::cvmath::Matrix<CV_64F> translation, rotation;

extern point pp[8];

typedef struct _error
{
	double adj_x, adj_y, adj_z;
	double posit_x, posit_y;
}Error;

void init();

void readAndConvert(
	opencv::cvmath::Matrix<CV_64F> &, opencv::cvmath::Matrix<CV_64F> &);
void readAndConvert2(
	opencv::cvmath::Matrix<CV_64F> &, opencv::cvmath::Matrix<CV_64F> &);
void findExtrinsic(const opencv::cvutil::Image<IPL_DEPTH_8U> &image, 
	const opencv::cvmath::Matrix<CV_64F> &intrinsic, const opencv::cvmath::Matrix<CV_64F> &distortion,
	opencv::cvmath::Matrix<CV_64F> &translation, opencv::cvmath::Matrix<CV_64F> &rotation);

void WtoC(
	const opencv::cvmath::Vector<CV_64F> &pW, const opencv::cvmath::Vector<CV_64F> &pcW,
	const opencv::cvmath::Matrix<CV_64F> &RcW, opencv::cvmath::Vector<CV_64F> &pI);
void CtoI(
	const opencv::cvmath::Vector<CV_64F> &pC, const opencv::cvmath::Matrix<CV_64F> &intrinsic,
	opencv::cvmath::Vector<CV_64F> &pI);
void CtoW(
	const opencv::cvmath::Vector<CV_64F> &pC, const opencv::cvmath::Vector<CV_64F> &pcW,
	const opencv::cvmath::Matrix<CV_64F> &RcW, opencv::cvmath::Vector<CV_64F> &pW);

void CalcRPY(
	const opencv::cvmath::Matrix<CV_64F> &RC,
	const opencv::cvmath::Matrix<CV_64F> &RoWC,
	opencv::cvmath::Vector<CV_64F> &thetaW);

void pCtopW(
	const opencv::cvmath::Vector<CV_64F> &positionC,
	const opencv::cvmath::Vector<CV_64F> &oWC,	// [h_̃JWnł̈ʒuxNg
	const opencv::cvmath::Matrix<CV_64F> &RoWC,	// [h_̃JWnł̉]s
	opencv::cvmath::Vector<CV_64F> &positionW);

void StoC(
	const opencv::cvmath::Vector<CV_64F> &pS_m, opencv::cvmath::Vector<CV_64F> &pC_mm);
void StoC(
	const opencv::cvmath::Vector<CV_64F> &positionS_m, opencv::cvmath::Vector<CV_64F> &positionC_mm,
	const opencv::cvmath::Vector<CV_64F> &thetaS_rad, opencv::cvmath::Vector<CV_64F> &thetaC_rad);

void pStopC(
	const opencv::cvmath::Vector<CV_64F> &positionS_m, opencv::cvmath::Vector<CV_64F> &positionC_mm);

void ErrorInit(Error (*ErrorMap)[35], int height, int width);

extern double calcdist(double x1, double y1, double x2, double y2);

int conv_zahyou(parameter *object_pose)
{
	opencv::cvutil::Image<IPL_DEPTH_8U> src_img(IMG_PATH, CV_LOAD_IMAGE_COLOR);
	opencv::cvutil::Image<IPL_DEPTH_8U> dst_img(src_img.getWidth(), src_img.getHeight(), src_img.getNChannels());

	parameter sensor_pose = *object_pose;

	std::cout << "isParameterLoaded" << isParameterLoaded << std::endl;

	if(isParameterLoaded == false)
	{	
		printf("ParameterLoad");

		
		// p[^ǂݍ
		readAndConvert(intrinsic, distortion);
		// Op[^ǂݍ
		readAndConvert2(translation, rotation);
		isParameterLoaded = true;
	}
	cout << "intrinsics" << endl;		intrinsic.print();
	cout << "distortion" << endl;		distortion.print();
	cout << "translation" << endl;	translation.print();
	cout << "rotation" << endl;			rotation.print();
	cout << endl;

	// ixNg
	opencv::cvmath::Vector<CV_64F> trans(3);
	trans(0) = translation(0, 0);
	trans(1) = translation(0, 1);
	trans(2) = translation(0, 2);

	// ]s
	opencv::cvmath::Matrix<CV_64F> R(3, 3);
	cvRodrigues2(rotation, R);

	opencv::cvmath::Vector<CV_64F> pW(3);	
	opencv::cvmath::Vector<CV_64F> pC0(3), pC1(3), pC2(3), pC3(3); // JWn
	opencv::cvmath::Vector<CV_64F> pI0(2), pI1(2), pI2(2), pI3(2); // 摜Wn

	//p̓_̓ep
	opencv::cvmath::Vector<CV_64F> iptI0(2), iptI1(2), iptI2(2), iptI3(2), iptI4(2), iptI5(2), iptI6(2), iptI7(2);  

	// e_vZ
	cout << "Undistort" << endl;
	pW(0) = 0;	pW(1) = 0;	pW(2) = 0;
	WtoC(pW, trans, R, pC0);
	CtoI(pC0, intrinsic, pI0);
//	cout << "pC0:"; pC0.print();
//	cout << "pI0:"; pI0.print();

	pW(0) = 250;	pW(1) = 0;	pW(2) = 0;
	WtoC(pW, trans, R, pC1);
	CtoI(pC1, intrinsic, pI1);
//	cout << "pI1:"; pI1.print();

	pW(0) = 0;	pW(1) = 300;	pW(2) = 0;
	WtoC(pW, trans, R, pC2);
	CtoI(pC2, intrinsic, pI2);
//	cout << "pI2:"; pI2.print();

	pW(0) = 250;	pW(1) = 300;	pW(2) = 0;
	WtoC(pW, trans, R, pC3);
	CtoI(pC3, intrinsic, pI3);
//	cout << "pI3:"; pI3.print();

#if 0
	pC0.print();	pI0.print();
	CtoW(pC0, trans, R, intrinsic, pW);
	pW.print();
	cout << endl;

	pC1.print();	pI1.print();
	CtoW(pC1, trans, R, intrinsic, pW);
	pW.print();
	cout << endl;

	pC2.print();	pI2.print();
	CtoW(pC2, trans, R, intrinsic, pW);
	pW.print();
	cout << endl;

	pC3.print();	pI3.print();
	CtoW(pC3, trans, R, intrinsic, pW);
	pW.print();
	cout << endl;
#endif

	/* cݕ␳O摜ւ̓e_vZ (E3D -> 摜2D) */
	opencv::cvmath::Matrix<CV_32F> p(4, 3), pI(4, 2);
	p(0, 0) = 0;		p(0, 1) = 0;		p(0, 2) = 0;
	p(1, 0) = 250;	p(1, 1) = 0;		p(1, 2) = 0;
	p(2, 0) = 0;		p(2, 1) = 300;	p(2, 2) = 0;
	p(3, 0) = 250;	p(3, 1) = 300;	p(3, 2) = 0;

	cvProjectPoints2(p, rotation, translation, intrinsic, distortion, pI);

		cout << "Distort" << endl;
//	pI.print();

	//Ԃ邽ߖ
	// 
	cvCircle(src_img, cvPoint(int(intrinsic(0,2)+0.5), int(intrinsic(1,2)+0.5)), 3, CV_RGB(255, 0, 0), CV_FILLED);


	/* cݕ␳ */
	// c݃}bv
	opencv::cvutil::Image<IPL_DEPTH_32F> mapx(src_img.getWidth(), src_img.getHeight(), 1);
	opencv::cvutil::Image<IPL_DEPTH_32F> mapy(src_img.getWidth(), src_img.getHeight(), 1);
	cvInitUndistortMap (intrinsic, distortion, mapx, mapy);

	// cݕ␳
	cvRemap (src_img, dst_img, mapx, mapy);

	/* e */
	cvCircle(dst_img, cvPoint(int(pI0(0)+0.5), int(pI0(1)+0.5)), 3, CV_RGB(255, 0, 0), CV_FILLED);
	cvCircle(dst_img, cvPoint(int(pI1(0)+0.5), int(pI1(1)+0.5)), 3, CV_RGB(0, 255, 0), CV_FILLED);
	cvCircle(dst_img, cvPoint(int(pI2(0)+0.5), int(pI2(1)+0.5)), 3, CV_RGB(0, 0, 255), CV_FILLED);
	cvCircle(dst_img, cvPoint(int(pI3(0)+0.5), int(pI3(1)+0.5)), 3, CV_RGB(255, 255, 0), CV_FILLED);

	cvCircle(src_img, cvPoint(int(pI(0,0)), int(pI(0,1))), 5, CV_RGB(255, 0, 0), CV_FILLED);
	cvCircle(src_img, cvPoint(int(pI(1,0)), int(pI(1,1))), 5, CV_RGB(0, 255, 0), CV_FILLED);
	cvCircle(src_img, cvPoint(int(pI(2,0)), int(pI(2,1))), 5, CV_RGB(0, 0, 255), CV_FILLED);
	cvCircle(src_img, cvPoint(int(pI(3,0)), int(pI(3,1))), 5, CV_RGB(255, 255, 0), CV_FILLED);

#if 0
	/**/
	opencv::cvmath::Vector<CV_64F> pS(3), pC(3);
	pW(0) = 150;	pW(1) = 150; pW(2) = 242;
	WtoC(pW, trans, R, pC);	CtoI(pC, intrinsic, pI0);
	cvCircle(dst_img, cvPoint(int(pI0(0)+0.5), int(pI0(1)+0.5)), 3, CV_RGB(255, 128, 0), 1);
	cout << "pW:"; pW.print();
	cout << "pC:"; pC.print();
	cout << "pI:"; pI0.print();

	pW(0) = 150;	pW(1) = 150; pW(2) = 0;
	WtoC(pW, trans, R, pC);
	CtoI(pC, intrinsic, pI0);
	cvCircle(dst_img, cvPoint(int(pI0(0)+0.5), int(pI0(1)+0.5)), 3, CV_RGB(255, 128, 0), 1);
	cout << "pW:"; pW.print();
	cout << "pC:"; pC.print();
	cout << "pI:"; pI0.print();
#endif

	//܂œẽeXg

#if 1
	{
		opencv::cvmath::Vector<CV_64F> pS(3), pC(3);

		pS(0) = sensor_pose.x;	pS(1) = sensor_pose.y;	pS(2) = sensor_pose.z;
		StoC(pS, pC);

		intrinsic(0,2) = src_img.getWidth()/2;
		intrinsic(1,2) = src_img.getHeight()/2;

		CtoW(pC, trans, R, pW);
		pW(2) -= 122;

		object_pose->x = pW(0);  object_pose->y = pW(1);  object_pose->z = pW(2);

		WtoC(pW, trans, R, pC);
		CtoI(pC, intrinsic, pI0);
		cvCircle(dst_img, cvPoint(int(pI0(0)+0.5), int(pI0(1)+0.5)), 3, CV_RGB(255, 128, 0), 1);
	}
#endif

//F̓_ɂ鐄茋ʂ̏o͂

	intrinsic(0,2) = src_img.getWidth()/2;
	intrinsic(1,2) = src_img.getHeight()/2;

	// f
	// p^CWOpp[c̃TCYɕς
	opencv::cvmath::Vector<CV_64F> fS[4], fC[4], RfC[4];
	//pp̓_
	opencv::cvmath::Vector<CV_64F> iptS[8], iptC[8], RiptC[8];

	fS[0].resize(3);			fS[0](0) = -0.02;	fS[0](1) = -0.015;	fS[0](2) = -0.025;
	fS[1].resize(3);			fS[1](0) = -0.02;	fS[1](1) = +0.015;	fS[1](2) = -0.025;
	fS[2].resize(3);			fS[2](0) = +0.02;	fS[2](1) = +0.015;	fS[2](2) = -0.025;
	fS[3].resize(3);			fS[3](0) = +0.02;	fS[3](1) = -0.015;	fS[3](2) = -0.025;

	fC[0].resize(3);			fC[1].resize(3);			fC[2].resize(3);		fC[3].resize(3);
	RfC[0].resize(3);			RfC[1].resize(3);			RfC[2].resize(3);		RfC[3].resize(3);

	StoC(fS[0], fC[0]);		StoC(fS[1], fC[1]);		StoC(fS[2], fC[2]);		StoC(fS[3], fC[3]);

	//pp̓_Wϊ

	iptS[0].resize(3);			iptS[0](0) = pp[0].x; iptS[0](1) = pp[0].y; iptS[0](2) = pp[0].z;
	iptS[1].resize(3);			iptS[1](0) = pp[1].x; iptS[1](1) = pp[1].y; iptS[1](2) = pp[1].z;
	iptS[2].resize(3);			iptS[2](0) = pp[2].x; iptS[2](1) = pp[2].y; iptS[2](2) = pp[2].z;
	iptS[3].resize(3);			iptS[3](0) = pp[3].x; iptS[3](1) = pp[3].y; iptS[3](2) = pp[3].z;
	iptS[4].resize(3);			iptS[4](0) = pp[4].x; iptS[4](1) = pp[4].y; iptS[4](2) = pp[4].z;
	iptS[5].resize(3);			iptS[5](0) = pp[5].x; iptS[5](1) = pp[5].y; iptS[5](2) = pp[5].z;
	iptS[6].resize(3);			iptS[6](0) = pp[6].x; iptS[6](1) = pp[6].y; iptS[6](2) = pp[6].z;
	iptS[7].resize(3);			iptS[7](0) = pp[7].x; iptS[7](1) = pp[7].y; iptS[7](2) = pp[7].z;

	iptC[0].resize(3); iptC[1].resize(3); iptC[2].resize(3); iptC[3].resize(3);
	iptC[4].resize(3); iptC[5].resize(3); iptC[6].resize(3); iptC[7].resize(3);

	RiptC[0].resize(3); RiptC[1].resize(3); RiptC[2].resize(3); RiptC[3].resize(3);
	RiptC[4].resize(3); RiptC[5].resize(3); RiptC[6].resize(3); RiptC[7].resize(3);

	StoC(iptS[0], iptC[0]); 	StoC(iptS[1], iptC[1]); 	StoC(iptS[2], iptC[2]); 	StoC(iptS[3], iptC[3]);
	StoC(iptS[4], iptC[4]); 	StoC(iptS[5], iptC[5]); 	StoC(iptS[6], iptC[6]); 	StoC(iptS[7], iptC[7]);


	opencv::cvmath::Vector<CV_64F> pS(3), thetaS(3);
	opencv::cvmath::Vector<CV_64F> pC(3), thetaC(3), axisC(3);

#if 1
	/* [OpenCVJWn] f̕i */
//	pS(0) = 0.003743;	pS(1) = 0.064221;	pS(2) = 0.660191;
	pS(0) = sensor_pose.x;	pS(1) = sensor_pose.y;	pS(2) = sensor_pose.z;
	StoC(pS, pC);

	/* [OpenCVJWn] f̉] */
//	thetaS(0) = -2.747216;	thetaS(1) = +0.003663;	thetaS(2) = +0.769445;
	thetaS(0) = sensor_pose.tx;	thetaS(1) = sensor_pose.ty;	thetaS(2) = sensor_pose.tz;

	thetaC(0) =  -thetaS(0);
	thetaC(1) =  -thetaS(1);
	thetaC(2) =  thetaS(2);

#else
	pS(0) = sensor_pose.x;	pS(1) = sensor_pose.y;	pS(2) = sensor_pose.z;
	thetaS(0) = sensor_pose.tx;	thetaS(1) = sensor_pose.ty;	thetaS(2) = sensor_pose.tz;

	thetaS.print();

	//SensorToCamera
	StoC(pS, pC, thetaS, thetaC);
//	printf("SensorToCamera\n");
//	printf("Sensor\n");
//	pS.print();
//	thetaS.print();
//	printf("Camera\n");
//	pC.print();
//	thetaC.print();
#endif
	/* loRŉ]s */
	opencv::cvmath::Quaternion<CV_64F> qXC, qYC, qZC; // ẻ]l
	opencv::cvmath::Quaternion<CV_64F> qC, qiC;
	

	axisC.setZero();	axisC(0) = 1; axisC.xS(thetaC(0));		qXC.fromAxis(axisC);
	axisC.setZero();	axisC(1) = 1; axisC.xS(thetaC(1));		qYC.fromAxis(axisC);
	axisC.setZero();	axisC(2) = 1; axisC.xS(thetaC(2));		qZC.fromAxis(axisC);

//	qC.QxQ(qZC, qYC);  qC.QxQ(qC, qXC);
	qC.QxQ(qXC, qYC);  qC.QxQ(qC, qZC);
	// ẻ]̍

#if 1
	opencv::cvmath::Matrix<CV_64F> RC(3, 3);
	qC.toR(RC);	// ]s

	//qC:JWł̎p

	/* e */
	// [OpenCVJWn] f]
	RfC[0].MxV(RC, fC[0]);	    RfC[1].MxV(RC, fC[1]);			RfC[2].MxV(RC, fC[2]);			RfC[3].MxV(RC, fC[3]);
	
	// [OpenCVJWn] fiړ
	RfC[0].VplusV(RfC[0], pC);	RfC[1].VplusV(RfC[1], pC);	RfC[2].VplusV(RfC[2], pC);	RfC[3].VplusV(RfC[3], pC);

	// e(eXgp)
	CtoI(RfC[0], intrinsic, pI0);			CtoI(RfC[1], intrinsic, pI1);			CtoI(RfC[2], intrinsic, pI2);
	CtoI(RfC[3], intrinsic, pI3);		

#endif


	//덷}bv̐錾

	Error ErrorMap[35];

	ErrorInit(&ErrorMap, 5, 7);

	//덷xNg`
	opencv::cvmath::Vector<CV_64F> errorC(3);	

	double min_dist = 9999.9;
	double new_dist;
	int min_ind = 0;

	for(int i = 0; i < 35; i++)
	{
		new_dist = calcdist(ErrorMap[i].posit_x,  ErrorMap[i].posit_y ,pC(0), pC(1));
		if(min_dist > new_dist)
		{
			min_dist = new_dist;
			min_ind = i;
		}
	}

	errorC(0) = ErrorMap[min_ind].adj_x;
	errorC(1) = ErrorMap[min_ind].adj_y;
	errorC(2) = ErrorMap[min_ind].adj_z;

	printf("error_vector_Camera");
	errorC.print();
	printf("ok?\n");	

	//JWnɂ錋ʂ덷ϊ
	opencv::cvmath::Vector<CV_64F> adj_pC(3);
	opencv::cvmath::Vector<CV_64F> adj_pW(3);

	adj_pC(0) = pC(0) + errorC(0);
	adj_pC(1) = pC(1) + errorC(1);
	adj_pC(2) = pC(2) + errorC(2);

	opencv::cvmath::Vector<CV_64F> posW(3);
	opencv::cvmath::Vector<CV_64F> thetaW(3);

	//C̐茋ʂWϊ
	pCtopW(pC, trans, R, posW);
	CalcRPY(RC, R, thetaW);

	//C̐茋ʂWϊ
	pCtopW(adj_pC, trans, R, adj_pW);

	//킩₷邽߂degreeɕϊ(0929)

	object_pose->tx = opencv::cvmath::rad2deg(thetaW(0));	
	object_pose->ty = opencv::cvmath::rad2deg(thetaW(1));	
	object_pose->tz = opencv::cvmath::rad2deg(thetaW(2));

	printf("Result of PoseEstimation(in Camera Coordinate System)--\n");
	pC.print();
	printf("tx:%3.3f(deg), ty:%3.3f(deg), tz:%3.3f(deg)\n",opencv::cvmath::rad2deg(thetaC(0)), opencv::cvmath::rad2deg(thetaC(1)), opencv::cvmath::rad2deg(thetaC(2)));
	printf("-------------------------------------------------------\n");

	printf("Result of PoseEstimation(in World Coordinate System)---\n");
	posW.print();
	printf("tx:%3.3f(deg), ty:%3.3f(deg), tz:%3.3f(deg)\n",object_pose->tx, object_pose->ty, object_pose->tz);
	printf("-------------------------------------------------------\n");

	printf("Adjusted Result of PoseEstimation(in Camera Coordinate System)--\n");
	adj_pC.print();
	printf("-------------------------------------------------------\n");	

	printf("Adjusted Result of PoseEstimation(in World Coordinate System)---\n");
	adj_pW.print();
	printf("-------------------------------------------------------\n");


	object_pose->x = adj_pW(0);		object_pose->y = adj_pW(1);		object_pose->z = adj_pW(2);
	object_pose->tx = thetaW(0);	object_pose->ty = thetaW(1);	object_pose->tz = thetaW(2);

	opencv::cvmath::Matrix<CV_64F> RW(3, 3);
	RW.MxM(R, RC);

#if 0
	/* [OpenCVEWn] ]s */
	opencv::cvmath::Matrix<CV_64F> RW(3, 3);
	RW.MxM(R, RC);
	opencv::cvmath::Vector<CV_64F> v[3], Rv[3];
	v[0].resize(3);		v[1].resize(3);		v[2].resize(3);
	Rv[0].resize(3);	Rv[1].resize(3);	Rv[2].resize(3);
	v[0].setZero();
	v[1].setZero();	v[1](0) = 100;
	v[2].setZero();	v[2](2) = 100;

	Rv[0].MxV(RW, v[0]);	Rv[1].MxV(RW, v[1]);	Rv[2].MxV(RW, v[2]);
	Rv[0].print();				Rv[1].print();				Rv[2].print();

	cout << "theta X = " << opencv::cvmath::rad2deg(atan2(Rv[2](2), Rv[2](1))) << endl;
	//cout << "theta Y = " << opencv::cvmath::rad2deg(atan2(Rv[0](2), Rv[0](1))) << endl;
	//cout << "theta Z = " << opencv::cvmath::rad2deg(atan2(Rv[0](2), Rv[0](1))) << endl;
#endif

	return 0;
}

int conv_zahyou_p(parameter *object_pose)
{
	parameter sensor_pose = *object_pose;


	if(isParameterLoaded == false)
	{	// p[^ǂݍ
		readAndConvert(intrinsic, distortion);

		readAndConvert2(translation, rotation);
		isParameterLoaded = true;
	}
	// ixNg
	opencv::cvmath::Vector<CV_64F> trans(3);
	trans(0) = translation(0, 0);
	trans(1) = translation(0, 1);
	trans(2) = translation(0, 2);

	// ]s
	opencv::cvmath::Matrix<CV_64F> R(3, 3);
	cvRodrigues2(rotation, R);

	opencv::cvmath::Vector<CV_64F> pS(3);
	opencv::cvmath::Vector<CV_64F> pC(3);


	pS(0) = sensor_pose.x;	pS(1) = sensor_pose.y;	pS(2) = sensor_pose.z;

	//SensorToCamera
	StoC(pS, pC);

	//VCtoW
	opencv::cvmath::Vector<CV_64F> posW(3);

	pCtopW(pC, trans, R, posW);

	object_pose->x = posW(0);	object_pose->y = posW(1);	object_pose->z = posW(2);

	return 0;
}


void readAndConvert(
	opencv::cvmath::Matrix<CV_64F> &intrinsic, opencv::cvmath::Matrix<CV_64F> &distortion)
{

	// sǂݍ / ϊ
	CvMat *_intrinsic, *_distortion;

	_intrinsic = (CvMat *)cvLoad("Intrinsics.xml");
	_distortion = (CvMat *)cvLoad("Distortion.xml");

	intrinsic.resize(_intrinsic->rows, _intrinsic->cols);
	for(int r = 0; r < _intrinsic->rows; ++ r)
		for(int c = 0; c < _intrinsic->cols; ++ c)
			intrinsic(r, c) = cvGetReal2D(_intrinsic, r, c);

	distortion.resize(_distortion->rows, _distortion->cols);
	for(int r = 0; r < _distortion->rows; ++ r)
		for(int c = 0; c < _distortion->cols; ++ c)
			distortion(r, c) = cvGetReal2D(_distortion, r, c);

	cvReleaseMat (&_intrinsic);
	cvReleaseMat (&_distortion);

}

void readAndConvert2(
	opencv::cvmath::Matrix<CV_64F> &translation, opencv::cvmath::Matrix<CV_64F> &rotation)
{
	// sǂݍ / ϊ
	CvMat *_translation, *_rotation;

	_translation = (CvMat *)cvLoad("Translation.xml");
	_rotation = (CvMat *)cvLoad("Rotation.xml");

	translation.resize(_translation->rows, _translation->cols);
	for(int r = 0; r < _translation->rows; ++ r)
		for(int c = 0; c < _translation->cols; ++ c)
			translation(r, c) = cvGetReal2D(_translation, r, c);

	rotation.resize(_rotation->rows, _rotation->cols);
	for(int r = 0; r < _rotation->rows; ++ r)
		for(int c = 0; c < _rotation->cols; ++ c)
			rotation(r, c) = cvGetReal2D(_rotation, r, c);

	cvReleaseMat (&_translation);
	cvReleaseMat (&_rotation);

}

/*
 * Op[^̌vZ
 */
void findExtrinsic(const opencv::cvutil::Image<IPL_DEPTH_8U> &image,
	const opencv::cvmath::Matrix<CV_64F> &intrinsic, const opencv::cvmath::Matrix<CV_64F> &distortion,
	opencv::cvmath::Matrix<CV_64F> &translation, opencv::cvmath::Matrix<CV_64F> &rotation)
{
#define IMAGE_NUM  (1)					/* 摜 */
#define PAT_ROW    (6)					/* p^[̍s */
#define PAT_COL    (7)					/* p^[̗ */
#define PAT_SIZE   (PAT_ROW*PAT_COL)
#define ALL_POINTS (IMAGE_NUM*PAT_SIZE)
#define CHESS_SIZE (50.0)				/* p^[1}X1ӃTCY[mm] */
	
	int i, j, k;
	int corner_count, found;

	CvSize pattern_size = cvSize (PAT_COL, PAT_ROW);
	CvPoint3D32f objects[ALL_POINTS];
	CvPoint2D32f *corners = (CvPoint2D32f *) cvAlloc (sizeof (CvPoint2D32f) * ALL_POINTS);
	CvMat object_points;
	CvMat image_points;
	
	CvMat *_intrinsic = cvCreateMat (3, 3, CV_32FC1);
	CvMat *_distortion = cvCreateMat (1, 4, CV_32FC1);

	CvMat *_rotation = cvCreateMat (1, 3, CV_32FC1);
	CvMat *_translation = cvCreateMat (1, 3, CV_32FC1);
	
	_intrinsic->data.fl[0] = float(intrinsic(0,0));	_intrinsic->data.fl[1] = float(intrinsic(0,1));	_intrinsic->data.fl[2] = float(intrinsic(0,2));
	_intrinsic->data.fl[3] = float(intrinsic(1,0));	_intrinsic->data.fl[4] = float(intrinsic(1,1));	_intrinsic->data.fl[5] = float(intrinsic(1,2));
	_intrinsic->data.fl[6] = float(intrinsic(2,0));	_intrinsic->data.fl[7] = float(intrinsic(2,1));	_intrinsic->data.fl[8] = float(intrinsic(2,2));

	_distortion->data.fl[0] = float(distortion(0,0));	_distortion->data.fl[1] = float(distortion(0,1));
	_distortion->data.fl[2] = float(distortion(0,2));	_distortion->data.fl[3] = float(distortion(0,3));

	// (2)3ԍW̐ݒ
	for (i = 0; i < IMAGE_NUM; i++) {
		for (j = 0; j < PAT_ROW; j++) {
			for (k = 0; k < PAT_COL; k++) {
				objects[i * PAT_SIZE + j * PAT_COL + k].x = float(j * CHESS_SIZE);
				objects[i * PAT_SIZE + j * PAT_COL + k].y = float(k * CHESS_SIZE);
				objects[i * PAT_SIZE + j * PAT_COL + k].z = 0.0;
			}
		}
	}
	
	cvInitMatHeader (&object_points, ALL_POINTS, 3, CV_32FC1, objects);
	
	// (3)`FX{[hiLu[Vp^[j̃R[i[o
	int found_num = 0;

	cvNamedWindow ("Calibration", CV_WINDOW_AUTOSIZE);

	found = cvFindChessboardCorners (image, pattern_size, corners, &corner_count);
	fprintf (stderr, "%02d...", i);
	if (found) {
		fprintf (stderr, "ok\n");
		found_num++;
	}
	else {
		fprintf (stderr, "fail\n");
	}
		
	// (4)R[i[ʒuTusNZxɏCC`
	opencv::cvutil::Image<IPL_DEPTH_8U> image_col(image.getWidth(), image.getHeight(), 3);
	image_col.copy(image);

	opencv::cvutil::Image<IPL_DEPTH_8U> image_gray(image_col.getWidth(), image_col.getHeight(), 1);
	cvCvtColor (image_col, image_gray, CV_BGR2GRAY);
		
	cvFindCornerSubPix (image_gray, corners, corner_count,
		cvSize (3, 3), cvSize (-1, -1), cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, 0.03));
		
	cvDrawChessboardCorners (image_col, pattern_size, corners, corner_count, found);

	cvShowImage ("Calibration", image_col);
	cvWaitKey (0);

	printf("%d\n", corner_count);

	cvDestroyWindow ("Calibration");
	
	if (found_num != IMAGE_NUM)
		return;
	
	translation.resize(1, 3);
	rotation.resize(1, 3);

	cvInitMatHeader (&image_points, ALL_POINTS, 1, CV_32FC2, corners);
	cvFindExtrinsicCameraParams2 (&object_points, &image_points, intrinsic, distortion, rotation, translation);

}


void WtoC(
	const opencv::cvmath::Vector<CV_64F> &pW, const opencv::cvmath::Vector<CV_64F> &oWC,
	const opencv::cvmath::Matrix<CV_64F> &RoWC, opencv::cvmath::Vector<CV_64F> &pC)
{
#if 0
	// [EWn] _p̈ʒuxNgCJ̈ʒuxNgCJʒu_pւ̃xNg
	opencv::cvmath::Vector<CV_64F> pW_pcW(3);
	pW_pcW.VminusV(pW, pcW);
	cout << "pW_pcW";	pW_pcW.print();
	
	opencv::cvmath::Matrix<CV_64F> Ri(3, 3);
	opencv::cvmath::Vector<CV_64F> pC(3);
	RcW.invert(Ri);

	pC.MxV(RcW, pW_pcW);
	cout << "    pC";	pC.print();
#else
	pC.MxV(RoWC, pW);
	pC(0) += oWC(0);	pC(1) += oWC(1);	pC(2) += oWC(2);
#endif
}
/*
 * 3W摜W֕ϊ
 */
void CtoI(
	const opencv::cvmath::Vector<CV_64F> &pC, const opencv::cvmath::Matrix<CV_64F> &intrinsic,
	opencv::cvmath::Vector<CV_64F> &pI)
{
	pI(0) = intrinsic(0, 0) * pC(0) / pC(2) + intrinsic(0, 2);
	pI(1) = intrinsic(1, 1) * pC(1) / pC(2) + intrinsic(1, 2);
}

/*
 * JWn琢EWn֕ϊ
 */
void CtoW(
	const opencv::cvmath::Vector<CV_64F> &pC,
	const opencv::cvmath::Vector<CV_64F> &oWC,	// [h_̃JWnł̈ʒuxNg
	const opencv::cvmath::Matrix<CV_64F> &RoWC,	// [h_̃JWnł̉]s
	opencv::cvmath::Vector<CV_64F> &pW)
{
	opencv::cvmath::Vector<CV_64F> pC_oWC(3);

	pC_oWC(0) = pC(0) - oWC(0);
	pC_oWC(1) = pC(1) - oWC(1);
	pC_oWC(2) = pC(2) - oWC(2);

	opencv::cvmath::Matrix<CV_64F> RoCW(3, 3);

	RoWC.invert(RoCW);

	pW.MxV(RoCW, pC_oWC);
}

/*
 * ]s񂩂ʒupp[^vZ
 */
void CalcRPY(
	const opencv::cvmath::Matrix<CV_64F> &RC,
	const opencv::cvmath::Matrix<CV_64F> &RoWC,	// [h_̃JWnł̉]s
	opencv::cvmath::Vector<CV_64F> &thetaW)
{
	opencv::cvmath::Matrix<CV_64F> RoWRC(3,3);
	opencv::cvmath::Matrix<CV_64F> RoCW(3,3);

	double tx, ty, tz;

	double r23, r33, r11, r12, r13;

	RoWC.invert(RoCW);

	RoWRC.MxM(RoCW,RC);

	printf("RotationMatrix_in_zahyou");
	RoWRC.print();


	r23 = RoWRC.get(1,2);
	r33 = RoWRC.get(2,2);
	r11 = RoWRC.get(0,0);
	r12 = RoWRC.get(0,1);
	r13 = RoWRC.get(0,2);

	tx = atan2(-r23, r33);
	ty = atan2(r13, sqrt((r23 * r23) + (r33 * r33)));
	tz = atan2(-r12, r11);
 
	thetaW(0) = tx;
	thetaW(1) = ty;
	thetaW(2) = tz;
}

/*
 * JWn̓_𐢊EWn֕ϊ
 */
void pCtopW(
	const opencv::cvmath::Vector<CV_64F> &positionC,
	const opencv::cvmath::Vector<CV_64F> &oWC,	// [h_̃JWnł̈ʒuxNg
	const opencv::cvmath::Matrix<CV_64F> &RoWC,	// [h_̃JWnł̉]s
	opencv::cvmath::Vector<CV_64F> &positionW)
{
	
	opencv::cvmath::Vector<CV_64F> pC_oWC(3);

	pC_oWC(0) = positionC(0) - oWC(0);
	pC_oWC(1) = positionC(1) - oWC(1);
	pC_oWC(2) = positionC(2) - oWC(2);

	opencv::cvmath::Matrix<CV_64F> RoCW(3, 3);

	RoWC.invert(RoCW);

	positionW.MxV(RoCW, pC_oWC);
}

void StoC(
	const opencv::cvmath::Vector<CV_64F> &pS_m, opencv::cvmath::Vector<CV_64F> &pC_mm)
{
	pC_mm(0) = -1000.0 * pS_m(0);
	pC_mm(1) = -1000.0 * pS_m(1);
	pC_mm(2) =  1000.0 * pS_m(2);
}

/*
 * ZTWnJWn֕ϊ
 */
void StoC(
	const opencv::cvmath::Vector<CV_64F> &positionS_m, opencv::cvmath::Vector<CV_64F> &positionC_mm,
	const opencv::cvmath::Vector<CV_64F> &thetaS_rad, opencv::cvmath::Vector<CV_64F> &thetaC_rad)
{
#if 0
	opencv::cvmath::Vector axis<CV_64F>(3);
	axis(0) = 0; axis(1) = 0; axis(2) = 1;
	axis.xS(M_PI);
	opencv::cvmath::Quaternion<CV_64F> q(axis);
	opencv::cvmath::Matrix<CV_64F> Rq(3,3);
	q.toR(Rq);

	pC_mm.MxV(Rq, pS_m);
	pC_mm.xS(1000.0);
#else

	positionC_mm(0) = -1000.0 * positionS_m(0);
	positionC_mm(1) = -1000.0 * positionS_m(1);
	positionC_mm(2) =  1000.0 * positionS_m(2);

	thetaC_rad(0) = -thetaS_rad(0);
	thetaC_rad(1) = -thetaS_rad(1);
	thetaC_rad(2) =  thetaS_rad(2);
#endif
}


void pStopC(
	const opencv::cvmath::Vector<CV_64F> &positionS_m, opencv::cvmath::Vector<CV_64F> &positionC_mm)
{
#if 0
	opencv::cvmath::Vector axis<CV_64F>(3);
	axis(0) = 0; axis(1) = 0; axis(2) = 1;
	axis.xS(M_PI);
	opencv::cvmath::Quaternion<CV_64F> q(axis);
	opencv::cvmath::Matrix<CV_64F> Rq(3,3);
	q.toR(Rq);

	pC_mm.MxV(Rq, pS_m);
	pC_mm.xS(1000.0);
#else

	positionC_mm(0) = -1000.0 * positionS_m(0);
	positionC_mm(1) = -1000.0 * positionS_m(1);
	positionC_mm(2) =  1000.0 * positionS_m(2);

#endif
}
/*
 * Wϊ
 */
void TransData()
{
	conv_zahyou(&object_pose);
}

/*
 * ZTWn̓_΍Wn֕ϊ
 */
CvPoint3D32f Trans3DPoints(double xval, double yval, double zval)
{
	parameter pointstatus;
	CvPoint3D32f returnpoint;
			
	pointstatus.x = xval;
	pointstatus.y = yval;
	pointstatus.z = zval;

	conv_zahyou_p(&pointstatus);

	returnpoint.x = pointstatus.x;
	returnpoint.y = pointstatus.y;
	returnpoint.z = pointstatus.z;

	return returnpoint;
}
/*
 * ZTWn̓_΍Wn֕ϊAt@C֏o
 * fobOp
 */
int Trans3DPointsandPrint(int pointnum, double *xval, double *yval, double *zval)
{
	parameter pointstatus;

	std::ofstream fout;
	fout.open("field_result.txt", std::ios::out);

	for(int i = 0; i < pointnum; i++)
	{
		pointstatus.x = xval[i];
		pointstatus.y = yval[i];
		pointstatus.z = zval[i];

		fout << pointstatus.x << ' ' << pointstatus.y << ' ' << pointstatus.z << '\n';

	}

	fout.close();
	return 0;
}
/*
 * G[}bvǂݍ
 */
void ErrorInit(Error (*ErrorMap)[35], int height, int width)
{
	//t@C
	//planeparameterǂݍ
	std::ifstream ifs("ErrorMap.txt");
	std::string str;
	char line[100], *ptr;
	double buf[5];
	int i, ind;

	if (ifs.fail())
	{
		fprintf(stderr, "Could not open file\n");
		return;
	}

	for(int h = 0; h < height; h++)
	{
		for(int w = 0; w < width; w++)
		{
			i = 0;

			if(ifs.getline(line, 100) == NULL)
				break;
		
			ptr = line;
			do{
				buf[i] = atof(ptr);
				ptr = strchr(ptr, ',');
				if(ptr!=NULL) {ptr++;}
				i++;
			}while(ptr!=NULL);

			ind = h * width + w;

			(*ErrorMap)[ind].adj_x = buf[0];
			(*ErrorMap)[ind].adj_y = buf[1];
			(*ErrorMap)[ind].adj_z = buf[2];
			(*ErrorMap)[ind].posit_x = buf[3];
			(*ErrorMap)[ind].posit_y = buf[4];
		}
	}
}
